#  Copyright (c) 2022 by Amplo.
from copy import deepcopy

import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    log_loss,
    max_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    r2_score,
    roc_auc_score,
)
from sklearn.model_selection import BaseCrossValidator, cross_val_predict

from amplo.base import BasePredictor, LoggingMixin

__all__ = ["ModelValidator"]


class ModelValidator(LoggingMixin):
    """
    Model validator.

    Parameters
    ----------
    cv_splits : int, default: 5
        Number of cross validation splits.
    cv_shuffle : bool, default: True
        Whether to shuffle the data for cross validation.
    verbose : {0, 1, 2}, default: 1
        Verbosity for logging.
    """

    def __init__(self, target: str, cv: BaseCrossValidator, verbose: int = 1):
        super().__init__(verbose=verbose)
        self.target = target
        self.cv = cv

    def validate(self, model: BasePredictor, data: pd.DataFrame, mode: str):
        """
        Validate model and return performance metrics.

        Parameters
        ----------
        model :
            Model to be validated.
        data : pd.DataFrame
            Training data for validation.
        mode : {"classification", "regression"}
            Model mode.

        Returns
        -------
        metrics : dict
            Cross validated model performance metrics.
        """
        model = deepcopy(model)
        y = data[self.target]
        x = data.drop(self.target, axis=1)

        # Calculate metrics
        if mode == "classification":
            metrics = self._validate_classification(model, x, y)
        elif mode == "regression":
            metrics = self._validate_regression(model, x, y)
        else:
            raise NotImplementedError("Invalid mode for validation.")

        # Logging
        for name, value in metrics.items():
            # We expect a mean and std value
            if not isinstance(value, float):
                continue
            name = f"{name.replace('_', ' ').title()}:".ljust(20)
            self.logger.info(f"{name} {value:.5f}")

        if mode == "classification":
            self.logger.info(
                """Confusion Matrix: 
            Predicted / actual  |    Positive    |    Negative    |
            Positive            |  {}  |  {}  |
            Negative            |  {}  |  {}  |
            """.format(
                    f"{metrics.get('true_positives')}".rjust(12),
                    f"{metrics.get('false_positives')}".rjust(12),
                    f"{metrics.get('false_negatives')}".rjust(12),
                    f"{metrics.get('true_negatives')}".rjust(12),
                )
            )

        return metrics

    def _validate_classification(self, model, x, y):
        """
        Cross validate classification (binary or multiclass) model.

        Parameters
        ----------
        model :
            Model to be validated.
        x : np.ndarray
            Training data for validation.
        y : np.ndarray
            Target data for validation.

        Returns
        -------
        metrics : dict
            Cross validated performance metrics of model.
        """
        labels = np.unique(y)
        log_loss_ = None

        # Modelling
        if hasattr(model, "predict_proba"):
            yp = cross_val_predict(model, x, y, cv=self.cv, method="predict_proba")
            log_loss_ = log_loss(y, yp)
            assert isinstance(yp, np.ndarray) and yp.shape == (len(y), len(labels))
            yp = np.argmax(yp, axis=1)
        else:
            yp = cross_val_predict(model, x, y, cv=self.cv)

        # Summarize metrics
        tp = np.sum(np.logical_and(y == 1, yp == 1))
        fp = np.sum(np.logical_and(y == 0, yp == 1))
        tn = np.sum(np.logical_and(y == 0, yp == 0))
        fn = np.sum(np.logical_and(y == 1, yp == 0))
        metrics = {
            "log_loss": log_loss_,
            "roc_auc": roc_auc_score(y, yp, labels=labels),
            "true_positives": tp,
            "false_positives": fp,
            "true_negatives": tn,
            "false_negatives": fn,
            "accuracy": accuracy_score(y, yp),
            "precision": tp / (tp + fp),
            "sensitivity_": tp / (tp + fn),
            "specificity": tn / (tn + fp),
            "f1_score": f1_score(y, yp, labels=labels, average="weighted"),
        }

        return metrics

    def _validate_regression(self, model, x, y):
        """
        Cross validate regression model.

        Parameters
        ----------
        model :
            Model to be validated.
        x : np.ndarray
            Training data for validation.
        y : np.ndarray
            Target data for validation.

        Returns
        -------
        metrics : dict
            Cross validated performance metrics of model.
        """
        # Modelling
        yp = cross_val_predict(model, x, y)

        # Summarize metrics
        metrics = {
            "mean_relative_error": mean_absolute_percentage_error(y, yp),
            "r2_score": r2_score(y, yp),
            "mean_absolute_error": mean_absolute_error(y, yp),
            "mean_squared_error": mean_squared_error(y, yp),
            "max_error": max_error(y, yp),
        }

        return metrics



#  Copyright (c) 2022 by Amplo.

import os
from typing import Union

import joblib
import numpy as np
import pandas as pd
import pytest
from sklearn.base import clone
from sklearn.datasets import make_classification, make_regression
from sklearn.metrics import get_scorer

from amplo.base.exceptions import NotFittedError
from amplo.classification import (
    CatBoostClassifier,
    LGBMClassifier,
    StackingClassifier,
    XGBClassifier,
)
from amplo.regression import (
    CatBoostRegressor,
    LGBMRegressor,
    StackingRegressor,
    XGBRegressor,
)

Model = Union[
    CatBoostClassifier,
    LGBMClassifier,
    XGBClassifier,
    StackingClassifier,
    CatBoostRegressor,
    LGBMRegressor,
    XGBRegressor,
    StackingRegressor,
]


setup_class_params = [
    CatBoostClassifier,
    LGBMClassifier,
    XGBClassifier,
    StackingClassifier,
    CatBoostRegressor,
    LGBMRegressor,
    XGBRegressor,
    StackingRegressor,
]


@pytest.fixture(scope="class", params=setup_class_params)
def setup_class(request):
    model = request.param

    if "Classifier" in model.__name__:
        x, y = make_classification(n_classes=5, n_informative=15)
        is_classification = True
    elif "Regressor" in model.__name__:
        x, y = make_regression(n_informative=15)
        is_classification = False
    else:
        raise ValueError("Invalid model requested")

    if "Stacking" in model.__name__:
        if is_classification:
            model_params = {"DecisionTreeClassifier__max_depth": 10}
        else:
            model_params = {"DecisionTreeRegressor__max_depth": 10}
    else:
        model_params = {"max_depth": 10}

    request.cls._model = model
    request.cls.model_params = model_params
    request.cls.is_classification = is_classification
    request.cls.x = pd.DataFrame(x)
    request.cls.y = pd.Series(y)
    yield


@pytest.mark.usefixtures("setup_class")
class TestModel:
    @pytest.fixture(autouse=True)
    def setup(self):
        # Initialize model
        self.model = self._model()
        yield

    def test_set_params(self):
        self.model.set_params(**self.model_params)

    def test_get_params(self):
        self.model.get_params()

    def test_fit_pandas(self):
        self.model.fit(self.x, self.y)

    def test_fit_numpy(self):
        self.model.fit(self.x.to_numpy(), self.y.to_numpy())

    def test_is_fitted(self):
        with pytest.raises(NotFittedError):
            self.model.check_is_fitted()
        self.model.fit(self.x, self.y)
        self.model.check_is_fitted()

    def test_predict(self):
        self.model.fit(self.x, self.y)
        prediction = self.model.predict(self.x)

        assert len(prediction.shape) == 1
        if self.is_classification:
            assert np.allclose(prediction.astype("int"), prediction)

    def test_predict_proba(self):
        if self.is_classification:
            self.model.fit(self.x, self.y)
            prediction = self.model.predict_proba(self.x)

            assert not np.isnan(prediction).any(), "NaN in prediction: {}".format(
                prediction
            )
            assert len(prediction.shape) == 2
            assert prediction.shape[1] == len(np.unique(self.y))
            assert np.allclose(np.sum(prediction, axis=1), 1)

    def test_cloneable(self):
        clone_model: Model = clone(self.model.set_params(**self.model_params))  # noqa
        clone_model.fit(self.x, self.y)

        if self.is_classification:
            prediction = clone_model.predict_proba(self.x)
            assert len(prediction.shape) != 1
            assert np.allclose(np.sum(prediction, axis=1), 1)
        else:
            prediction = clone_model.predict(self.x)
            assert len(prediction.shape) == 1

    def test_scorer(self):
        self.model.fit(self.x, self.y)

        if self.is_classification:
            scorers = ["neg_log_loss", "accuracy", "f1_micro"]
        else:
            scorers = ["neg_mean_squared_error", "neg_mean_absolute_error", "r2"]

        for name in scorers:
            get_scorer(name)(self.model, self.x, self.y)

    def test_binary(self):
        if self.is_classification:
            x, y = make_classification()
            self.model.fit(x, y)
            self.model.predict(x)
            self.model.predict_proba(x)
            get_scorer("neg_log_loss")(self.model, x, y)
            get_scorer("accuracy")(self.model, x, y)
            get_scorer("f1_micro")(self.model, x, y)

    def test_pickleable(self):
        x, y = make_classification()
        self.model.fit(x, y)
        joblib.dump(self.model, "tmp.joblib")
        os.remove("tmp.joblib")



#  Copyright (c) 2022 by Amplo.

import json

import pandas as pd
import pytest

from amplo.automl.feature_processing.nop_feature_extractor import NopFeatureExtractor


class TestNopFeatureExtractor:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_all(self, mode, make_data):
        data = make_data
        fe = NopFeatureExtractor(target="target")

        # Test fit and fit_transform
        out1 = fe.fit_transform(data)
        out2 = fe.transform(data)
        assert all(out1 == out2), "`fit_transform` and `transform` don't match."

        # Test features_
        features = set(data) - {"target"}
        assert set(fe.features_) == features, "Not all / too many features accepted."

        # Test settings
        new_fe = NopFeatureExtractor().load_settings(fe.get_settings())
        new_out = new_fe.transform(data)
        assert set(new_fe.features_) == features, "Features not correctly restored."
        assert all(out1 == new_out), "Transformation not correct."

        # Test JSON serializable
        settings = json.loads(json.dumps(fe.get_settings()))
        new_fe = NopFeatureExtractor().load_settings(settings)
        assert fe.get_settings() == new_fe.get_settings()
        assert all(fe.transform(data) == new_fe.transform(data))



#  Copyright (c) 2022 by Amplo.

from amplo.automl.data_processing import DataProcessor
from amplo.automl.drift_detection import DriftDetector
from amplo.automl.feature_processing import FeatureProcessor
from amplo.automl.grid_search import OptunaGridSearch
from amplo.automl.interval_analysis import IntervalAnalyser
from amplo.automl.modelling import Modeller
from amplo.automl.sequencing import Sequencer
from amplo.automl.standardization import Standardizer



#  Copyright (c) 2022 by Amplo.

import json

import numpy as np
import pandas as pd
import pytest

from amplo.automl.feature_processing.static_feature_extractor import (
    StaticFeatureExtractor,
)


@pytest.mark.usefixtures("make_rng")
class TestStaticFeatureExtractor:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_mode_and_settings(self, mode, make_x_y):
        x, y = make_x_y
        x = x.iloc[:, :5]  # for speed up
        data = pd.DataFrame(x).copy()
        data["target"] = y
        fe = StaticFeatureExtractor(target="target", mode=mode)

        # Test output
        out1 = fe.fit_transform(data)
        out2 = fe.transform(x)
        assert set(out1) == set(fe.features_), "`features_` doesn't match output."
        assert all(out1 == out2), "`fit_transform` and `transform` don't match."

        # Test settings
        new_fe = StaticFeatureExtractor().load_settings(fe.get_settings())
        assert fe.get_settings() == new_fe.get_settings()
        new_out = new_fe.transform(x)
        assert all(out1 == new_out), "FE loaded from settings has invalid output."
        assert set(fe.features_) == set(
            new_fe.features_
        ), "FE from settings has erroneous `features_`."

        # Test JSON serializable
        settings = json.loads(json.dumps(fe.get_settings()))
        new_fe = StaticFeatureExtractor().load_settings(settings)
        assert fe.get_settings() == new_fe.get_settings()
        assert all(fe.transform(x) == new_fe.transform(x))

    @pytest.mark.parametrize("mode", ["classification"])
    def test_raw_features(self, mode, make_x_y):
        x, y = make_x_y

        # Fit and check features
        fe = StaticFeatureExtractor(mode=mode)
        fe._fit_transform_raw_features(x, y)
        fe._is_fitted = True
        assert set(fe.features_) == set(fe.raw_features_)
        assert set(fe.features_) == set(x), "All columns should be accepted."

        # Test settings and transformation
        new_fe = StaticFeatureExtractor().load_settings(fe.get_settings())
        out = new_fe.transform(x)
        assert set(fe.features_) == set(out), "Expected columns don't match."

    @pytest.mark.parametrize("mode", ["regression"])
    def test_cross_features(self, mode):
        size = 100
        y = pd.Series(np.linspace(2, 100, size))
        random = pd.Series(self.rng.geometric(0.1, size))
        x = pd.DataFrame({"a": y / random, "b": y * random, "random": random})

        # Fit and check features
        fe = StaticFeatureExtractor(mode=mode)
        fe._set_validation_model()
        fe._init_feature_baseline_scores(x, y)
        fe._fit_transform_cross_features(x, y)
        fe._is_fitted = True
        assert set(fe.features_) == set(fe.cross_features_)
        assert "a__mul__random" in fe.features_, "Multiplicative feature not found."
        assert "b__div__random" in fe.features_, "Division feature not found."

        # Test settings and transformation
        new_fe = StaticFeatureExtractor().load_settings(fe.get_settings())
        out = new_fe.transform(x)
        assert set(fe.features_) == set(out), "Expected columns don't match."

    @pytest.mark.parametrize("mode", ["classification"])
    def test_k_means_features(self, mode):
        size = 100
        y = pd.Series([0] * (size // 2) + [1] * (size // 2))
        blob1 = self.rng.normal(0, 1, size // 2)
        blob2 = self.rng.normal(100, 1, size // 2)
        random = pd.Series(self.rng.geometric(0.1, size))
        x = pd.DataFrame({"blobs": [*blob1, *blob2], "random": random})

        # Fit and check features
        fe = StaticFeatureExtractor(mode=mode)
        fe._set_validation_model()
        fe._init_feature_baseline_scores(x, y)
        fe._fit_transform_k_means_features(x, y)
        fe._is_fitted = True
        assert set(fe.features_) == set(fe.k_means_features_)
        assert {"dist__0_2", "dist__1_2"}.issubset(
            fe.features_
        ), "K-Means features not found."

        # Test settings and transformation
        new_fe = StaticFeatureExtractor().load_settings(fe.get_settings())
        out = new_fe.transform(x)
        assert set(fe.features_) == set(out), "Expected columns don't match."

    @pytest.mark.parametrize("mode", ["regression"])
    def test_trigo_features(self, mode, make_x_y):
        size = 100
        y = pd.Series(self.rng.uniform(-1, 1, size=size))
        random = pd.Series(self.rng.geometric(0.1, size))
        x = pd.DataFrame(
            {"sinus": np.arcsin(y), "cosine": np.arccos(y), "random": random}
        )

        # Fit and check features
        fe = StaticFeatureExtractor(mode=mode)
        fe._set_validation_model()
        fe._init_feature_baseline_scores(x, y)
        fe._fit_transform_trigo_features(x, y)
        fe._is_fitted = True
        assert set(fe.features_) == set(fe.trigo_features_)
        assert "sin__sinus" in fe.features_, "Sinus feature not found."
        assert "cos__cosine" in fe.features_, "Cosine feature not found."

        # Test settings and transformation
        new_fe = StaticFeatureExtractor().load_settings(fe.get_settings())
        out = new_fe.transform(x)
        assert set(fe.features_) == set(out), "Expected columns don't match."

    @pytest.mark.parametrize("mode", ["regression"])
    def test_inverse_features(self, mode):
        size = 100
        y = pd.Series(self.rng.uniform(-1, 1, size=size))
        random = pd.Series(self.rng.geometric(0.1, size))
        x = pd.DataFrame({"inversed": (1.0 / y), "random": random})

        # Fit and check features
        fe = StaticFeatureExtractor(mode=mode)
        x = fe._check_x(x)  # sanitize
        fe._set_validation_model()
        fe._init_feature_baseline_scores(x, y)
        fe._fit_transform_inverse_features(x, y)
        fe._is_fitted = True
        assert set(fe.features_) == set(fe.inverse_features_)
        assert "inv__inversed" in fe.features_, "Inverse feature not found."

        # Test settings and transformation
        new_fe = StaticFeatureExtractor().load_settings(fe.get_settings())
        out = new_fe.transform(x)
        assert set(fe.features_) == set(out), "Expected columns don't match."



#  Copyright (c) 2022 by Amplo.

import time
import warnings

import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC, SVR

from amplo.classification import CatBoostClassifier
from amplo.observation._base import ProductionWarning
from amplo.observation.model import ModelObserver
from amplo.regression import CatBoostRegressor
from tests import (
    RandomPredictor,
    _OverfitClassifier,
    _OverfitRegressor,
    _RandomClassifier,
    _RandomRegressor,
)


@pytest.fixture
def make_one_to_one_data(mode):
    size = 100
    if mode == "classification":
        linear_col = np.random.choice([0, 1, 2], size)
    elif mode == "regression":
        linear_col = np.random.uniform(0.0, 1.0, size)
    else:
        raise ValueError("Invalid mode")
    yield pd.DataFrame({"x": linear_col, "target": linear_col})


class DelayedRandomPredictor(RandomPredictor):
    def __init__(self, delay: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delay = delay

    def predict(self, x):
        time.sleep(self.delay)
        return super().predict(x)

    def predict_proba(self, x):
        time.sleep(self.delay)
        return super().predict_proba(x)


class TestModelObserver:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_model_size(self, mode):
        with warnings.catch_warnings(record=True) as record:
            ModelObserver().check_model_size(RandomPredictor(mode=mode))
        assert not any(
            isinstance(r.message, ProductionWarning) for r in record
        ), "An unnecessary warning was raised while checking model size."

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_better_than_linear(self, mode, make_one_to_one_data):
        data: pd.DataFrame = make_one_to_one_data
        if mode == "classification":
            model = _RandomClassifier()
        else:
            model = _RandomRegressor()

        # Observe
        with pytest.warns(ProductionWarning):
            ModelObserver().check_better_than_linear(
                model=model, data=data, target="target", mode=mode
            )

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_noise_invariance(self, mode, make_data):
        if mode == "classification":
            model = _OverfitClassifier()
        if mode == "regression":
            model = _OverfitRegressor()

        data = make_data
        data = pd.concat([data, data], axis=0)
        with pytest.warns(ProductionWarning):
            ModelObserver().check_noise_invariance(
                model, data=data, target="target", mode=mode  # type: ignore
            )

        # Should not trigger normally
        ModelObserver().check_noise_invariance(
            model, data=data, target="target", mode=mode  # type: ignore
        )

    def test_check_slice_invariance(self):
        """
        This is a complex test. Slice invariance will be triggered with a linear
        model, when 90% of the data is linearly separable, but 10% is displaced
        compared to the fit.

        We just do this for classification for ease, the observer runs
        irrespective of mode.
        """
        # Classification dataset
        data = pd.DataFrame(
            {
                "x": np.concatenate((np.linspace(0, 90, 96), np.ones(4) * 100)),
                "target": np.concatenate(
                    (
                        np.zeros(48),
                        np.ones(48),
                        np.zeros(4),
                    )
                ),
            }
        )

        # Model
        model = LogisticRegression()
        model.fit(data["x"].values.reshape((-1, 1)), data["target"])  # type: ignore

        # Observe
        with pytest.warns(ProductionWarning):
            ModelObserver().check_slice_invariance(
                model=model,  # type: ignore
                data=data,
                target="target",
                mode="classification",
            )

        # Should not trigger normally
        ModelObserver().check_slice_invariance(
            model=_RandomClassifier(),
            data=data,
            target="target",
            mode="classification",
        )

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_boosting_overfit(self, mode, make_data):
        n_estimators = 1000
        boost_kwargs = dict(
            n_estimators=n_estimators,
            l2_leaf_reg=0,
            early_stopping_rounds=n_estimators,
            use_best_model=False,
        )
        data = make_data
        if mode == "classification":
            _model = CatBoostClassifier(**boost_kwargs)
        else:
            _model = CatBoostRegressor(**boost_kwargs)
        _model.fit(data.drop("target", axis=1), data["target"])

        # Observer
        with pytest.warns(ProductionWarning):
            ModelObserver().check_boosting_overfit(
                _model, data=data, target="target", mode=mode
            )



#  Copyright (c) 2022 by Amplo.

import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_classification, make_regression

from tests import rmfile, rmtree


@pytest.fixture(autouse=True)
def rmtree_automl():
    folder = "Auto_ML"
    rmtree(folder, must_exist=False)
    yield folder
    rmtree(folder, must_exist=False)


@pytest.fixture(autouse=True)
def rmfile_automl():
    file = "AutoML.log"
    yield file
    try:
        rmfile(file, must_exist=False)
    except PermissionError:
        pass


@pytest.fixture
@pytest.mark.parametrize("mode", ["classification", "regression"])
def make_x_y(request, mode):
    if mode == "classification":
        x, y = make_classification(n_features=5)
    elif mode == "multiclass":
        x, y = make_classification(n_features=5, n_classes=3, n_informative=3)
    elif mode == "regression":
        x, y = make_regression(n_features=5, noise=0.3)  # type: ignore
    else:
        raise ValueError("Invalid mode")
    x, y = pd.DataFrame(x), pd.Series(y)
    x.columns = [f"feature_{i}" for i in range(len(x.columns))]  # type: ignore
    y.name = "target"
    request.mode = mode
    yield x, y


@pytest.fixture
def make_data(request, make_x_y, target="target"):
    data, y = make_x_y
    data[target] = y
    request.data = data
    yield data


@pytest.fixture
def make_rng(request):
    request.cls.rng = np.random.default_rng(seed=92938)
    yield


@pytest.fixture(scope="class", params=["regression", "classification"])
def make_mode(request):
    mode = request.param
    target = "target"
    if mode == "classification":
        x, y = make_classification(n_features=5)
        request.cls.objective = "neg_log_loss"
    elif mode == "multiclass":
        x, y = make_classification(n_features=5, n_classes=3, n_informative=3)
        request.cls.objective = "neg_log_loss"
    elif mode == "regression":
        x, y = make_regression(n_features=5)  # type: ignore
        request.cls.objective = "neg_mean_squared_error"
    else:
        raise ValueError("Invalid mode")
    x, y = pd.DataFrame(x), pd.Series(y)
    x = x.rename({col: f"feature_{col}" for col in x.columns}, axis=1)
    request.cls.mode = mode
    request.cls.target = target
    request.cls.data = pd.concat([x, y.to_frame(target)], axis=1)
    yield



#  Copyright (c) 2022 by Amplo.

"""
Feature processor for extracting temporal features.
"""


from __future__ import annotations

import re
import time
from typing import Any, TypeVar, cast
from warnings import warn

import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl
import pywt
from scipy import signal
from tqdm import tqdm

from amplo.automl.feature_processing._base import (
    BaseFeatureExtractor,
    sanitize_dataframe,
)
from amplo.automl.feature_processing.pooling import get_pool_functions
from amplo.utils import check_dtypes

__all__ = [
    "pl_pool",
    "ScoreWatcher",
    "TemporalFeatureExtractor",
]

PandasType = TypeVar("PandasType", pd.Series, pd.DataFrame)


# ----------------------------------------------------------------------
# Pooling


def pl_pool(
    series: pd.Series,
    window_size: int,
    aggregation: dict[str, pl.Expr | Any],
    *,
    use_multi_index: bool = True,
) -> pd.DataFrame:
    """
    Pools series data with given aggregation functions.

    Parameters
    ----------
    series : pd.Series
        Data to be pooled.
    window_size : int
        Window size for pooling.
    aggregation : dict of {str: pl.Expr or typing.Any}
        Aggregation functions for pooling.
    use_multi_index : bool
        Whether to make use of multi-index if present.
        If False, pooling won't happen in a "groupby" fashion.

    Returns
    -------
    pd.DataFrame
        Pooled data where each column name consists of the original series data name and
        its pooling function name (keys of `aggregation` parameter).
        When for example the series data name is "series" and one `aggregation` key is
        named "min", the resulting column is named "series__pool=min".
    """
    # Integrity checks
    if series.name is None:
        series.name = "series"
    series.index.names = [
        name if isinstance(name, str) else f"index_{i}"
        for i, name in enumerate(series.index.names)
    ]

    # Make polars aggregator
    pl_agg = []
    for name, func in aggregation.items():
        # Check validity of `aggregation`
        if not hasattr(func, "__call__"):
            raise ValueError(f"Aggregator is not callable: {name}")

        # Convert function to polars
        if func.__module__ == "polars.internals.lazy_functions":
            func = func(str(series.name))  # type: ignore
        elif func.__module__ == "polars.internals.expr":
            func = func(pl.col(str(series.name)))  # type: ignore
        else:
            func = pl.col(str(series.name)).apply(func)  # type: ignore

        # Rename output variable and add to `pl_agg`
        func = func.alias(f"{series.name}__pool={name}")
        pl_agg.append(func)

    # Convert series to polars DataFrame
    pl_df = pl.DataFrame(series.reset_index(drop=False))

    # Set up pooling parameters
    is_double_index = len(series.index.names) == 2
    if not is_double_index or not use_multi_index:
        by, index_column = None, pl_df.columns[0]
    else:
        by, index_column = pl_df.columns[:2]

    # Make sure that MultiIndices of 3 or more don't get lost in `agg`
    if len(series.index.names) > 2:
        pl_agg += [pl.col(col).first() for col in series.index.names[2:]]

    # Pool and return
    return (
        pl_df.groupby_dynamic(index_column, every=f"{window_size}i", by=by)
        .agg(pl_agg)
        .to_pandas()
        .set_index(series.index.names)
    )


# ----------------------------------------------------------------------
# Wavelet extraction


def _extract_wavelets(series, scales, wavelet, name=None):
    check_dtypes("series", series, pd.Series)

    # Transform
    # Note that sampling_frequency does not matter.
    coeffs, _ = pywt.cwt(series, scales=scales, wavelet=wavelet)

    # Make dataframe
    columns = [f"{name or series.name}__wav__{wavelet}__{s}" for s in scales]
    x_out = pd.DataFrame(coeffs.real.T, index=series.index, columns=columns)

    return x_out


# ----------------------------------------------------------------------
# Main


class ScoreWatcher:
    """
    Watcher for scores.

    Parameters
    ----------
    keys : list of str
        Keys for the watcher.

    Attributes
    ----------
    watch : dict of {str: tuple of (int, np.ndarray)}
        Keeps track of the counter and score of each watcher key.
    """

    def __init__(self, keys: list[str]):
        check_dtypes("keys", keys, list)
        check_dtypes(("key__item", item, str) for item in keys)
        self.watch: dict[str, tuple[int, int, np.ndarray]] = {
            key: (0, 0, np.array(0)) for key in keys
        }

    def __getitem__(self, key: str) -> tuple[int, npt.NDArray[np.floating]]:
        """
        Get the counter and score for the given key.

        Parameters
        ----------
        key : str
            Key of the watcher.

        Returns
        -------
        typing.Tuple[int, np.ndarray]
        """
        count, weight, score = self.watch[key]
        return count, score

    def __repr__(self):
        """
        Readable string representation of the class.
        """
        return f"{self.__class__.__name__}({sorted(self.watch)})"

    def update(self, key: str, score: npt.ArrayLike, weight: int = 1) -> None:
        """
        Update a key of the watcher.

        Parameters
        ----------
        key : str
            Watcher key.
        score : array_like
            Scoring value(s).
        weight : int
            Weight of the score.

        Returns
        -------
        ScoreWatcher
            Updated instance of the watcher.
        """
        check_dtypes(("key", key, str), ("weight", weight, int))
        count, prev_weight, prev_score = self.watch[key]
        new_weight = prev_weight + weight
        new_score = prev_weight * prev_score + weight * np.asarray(score, float)
        if abs(new_weight) > 0:
            new_score /= new_weight  # type: ignore
        self.watch[key] = (count + 1, new_weight, np.asarray(new_score, float))

    def mean(self) -> npt.NDArray[np.floating]:
        """
        Calculate the mean of all scores.

        Returns
        -------
        np.ndarray
            Mean of all scores.
        """
        scores = [s for counter, _, s in self.watch.values() if counter > 0]
        return np.asarray(scores, float).mean(0)

    def std(self) -> npt.NDArray[np.floating]:
        """
        Calculate the standard deviation of all scores.

        Returns
        -------
        np.ndarray
            Standard deviation of all scores.
        """
        scores = [s for counter, _, s in self.watch.values() if counter > 0]
        return np.asarray(scores, float).std(0)


class TemporalFeatureExtractor(BaseFeatureExtractor):
    """
    Feature extractor for temporal data.

    Parameters
    ----------
    mode : {"notset", "classification", "regression"}, optional, default: "notset"
        Model mode.
    window_size : int, optional, default: None
        Determines how many data rows will be collected and summarized by pooling.
        If None, will determine a reasonable window size for the data at hand.
    fit_raw : bool, default: True
        Whether to include pooling from raw features to extract features.
    fit_wavelets : bool or list of str, default: True
        Whether to search for pooled wavelet features.
        If True, default wavelets are used for fitting, i.e.
        ["cmor1.5-1.0", "gaus4", "gaus7", "cgau2", "cgau6", "mexh"].
        Each string must be a valid wavelet name (see notes).
    pooling : list of str, optional, default: None
        Pooling functions to be used for feature extraction.
        If None, defaults will be used.
    strategy : {"exhaustive", "random", "smart"}, default: "smart"
        Fitting strategy for feature extraction.
        If "exhaustive", use brute-force method.
        If "random", iterates on randomly shuffled feature-wavelet combinations and
        performs pooling on a random subset of `self.pooling` until end of iterator or
        timeout is reached.
        If "smart", similar to "random" but (1) skips unpromising features or wavelets
        and (2) uses promising poolings only.
    timeout : int, default: 1800
        Timeout in seconds for fitting. Has no effect when `strategy` is "exhaustive".
    verbose : int, default: 0
        Verbosity for logger.

    Notes
    -----
    Valid ``wavelet`` parameters can be found via:
    >>> import pywt
    >>> pywt.wavelist()
    """

    _add_to_settings = ["window_size_", *BaseFeatureExtractor._add_to_settings]

    def __init__(
        self,
        target: str = "target",
        mode: str = "notset",
        window_size: int | None = None,
        fit_raw: bool = True,
        fit_wavelets: bool | list[str] = True,
        pooling: list[str] | None = None,
        strategy: str = "smart",
        timeout: int = 1800,
        verbose: int = 0,
    ):
        super().__init__(target=target, mode=mode, verbose=verbose)

        # Warnings
        if self.mode == "regression":
            # Some notes for implementing regression:
            #  - It does not make sense to pool features and target.
            #  - Wavelet transformations may still add some value.
            msg = (
                "TemporalFeatureExtractor is not ready for regression. "
                "Behavior probably won't meet your expectations!"
            )
            warn(msg, UserWarning)

        # Check inputs and set defaults
        check_dtypes(
            ("window_size", window_size, (type(None), int)),
            ("fit_raw", fit_raw, bool),
            ("fit_wavelets", fit_wavelets, (bool, list)),
            ("pooling", pooling, (type(None), list)),
            ("strategy", strategy, str),
            ("timeout", timeout, int),
        )
        if fit_wavelets is True:
            fit_wavelets = ["cmor1.5-1.0", "gaus4", "gaus7", "cgau2", "cgau6", "mexh"]
        elif fit_wavelets:  # if not True, must be an iterable
            check_dtypes(("fit_wavelets__item", item, str) for item in fit_wavelets)
        pooling = list(get_pool_functions(pooling))  # validate

        # Integrity checks
        if strategy not in ("exhaustive", "random", "smart"):
            raise ValueError(f"Invalid value for `strategy`: {strategy}")
        if timeout <= 0:
            raise ValueError(f"`timeout` must be strictly positive but got: {timeout}")
        if not any([fit_raw, fit_wavelets]):
            raise ValueError(
                "Disabling all fitting functions is useless. Enable at least one feature extractor."
            )

        # Set attributes
        self.window_size_ = window_size
        self.fit_raw = fit_raw
        self.fit_wavelets = fit_wavelets
        self.pooling = pooling
        self.strategy = strategy
        self.timeout = timeout

    def fit(self, data: pd.DataFrame) -> "TemporalFeatureExtractor":
        # We implement fit_transform because we anyhow transform the data. Therefore,
        # when using fit_transform we don't have to do redundant transformations.
        self.fit_transform(data)
        return self

    def fit_transform(self, data: pd.DataFrame) -> pd.DataFrame:
        self.logger.info("Fitting data.")
        self.reset()

        # Input checks
        data, _ = self._assert_multiindex(data)
        x, y = self._check_data(data)
        numeric_cols = [
            col for col, typ in zip(x, x.dtypes) if np.issubdtype(typ, np.number)
        ]
        if set(numeric_cols) != set(x):
            warn(
                "Handling non-numeric data is (currently) not supported. "
                "Corresponding columns will be ignored.",
                UserWarning,
            )
            x = x[numeric_cols]

        # Initialize fitting
        self._set_validation_model()
        self._set_window_size(x.index)
        x, y = self._fit_data_to_window_size(x, y)  # type: ignore
        x = cast(pd.DataFrame, x)  # type hint
        y = cast(pd.Series, y)  # type hint

        # Calculate baseline scores (w/o taking time into account)
        self._init_feature_baseline_scores(x, y)

        # Fit features
        y_pooled = self._pool_target(y)
        x_out = pd.concat(
            [
                self._fit_transform_raw_features(x, y_pooled, update_baseline=True),
                self._fit_transform_wav_features(x, y_pooled, update_baseline=True),
            ],
            axis=1,
        )
        # Ensure ordering of columns & sanitize
        x_out = x_out[self.features_]
        x_out = sanitize_dataframe(x_out)

        self._is_fitted = True
        return x_out

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        self.logger.info("Transforming data.")
        self.check_is_fitted()

        # Handle input
        data, got_multiindex = self._assert_multiindex(data, raise_on_single=False)
        x, _ = self._check_data(data, require_y=False)
        x = self._fit_data_to_window_size(x)

        # Apply transformations
        x_out = pd.concat(
            [
                self._transform_raw_features(x),
                self._transform_wav_features(x),
            ],
            axis=1,
        )
        # Ensure ordering of columns & sanitize
        x_out = x_out[self.features_]
        x_out = sanitize_dataframe(x_out)

        # Return
        if got_multiindex:
            return x_out
        else:
            return x_out.set_index(x_out.index.droplevel(0))

    def transform_target(self, y: pd.Series) -> pd.Series:
        self.check_is_fitted()

        # Handle input
        y, got_multiindex = self._assert_multiindex(y, raise_on_single=False)
        y = self._check_y(y, copy=False)
        y = self._fit_data_to_window_size(y)  # type: ignore

        # Transform
        y_out = self._pool_target(y)

        # Return
        if got_multiindex:
            return y_out
        else:
            return y_out.set_index(y_out.index.droplevel(0))

    # ----------------------------------------------------------------------
    # Feature processing

    @property
    def raw_features_(self):
        out = [str(c) for c in self.features_ if not re.search(".+__.+__pool=.+", c)]
        return sorted(out)

    def _fit_transform_raw_features(self, x, y_pooled, update_baseline=True):
        if not self.fit_raw:
            self.logger.info("Skipped fitting raw features.")
            dummy_y = pd.Series(np.zeros(len(x)), index=x.index, dtype=np.int32)
            return pd.DataFrame(index=self._pool_target(dummy_y).index)

        self.logger.info("Fitting raw features.")

        # Pool all features
        x_pooled = self._pool_features(x, drop_nan_columns=True)

        # Score and decide which features to accept
        scores = self.select_scores(
            x_pooled.apply(self._calc_feature_scores, y=y_pooled, axis=0),  # type: ignore
            best_n_per_class=50,
            update_baseline=update_baseline,
        )
        x_out = x_pooled[scores.columns]
        self.logger.info(f"Accepted {x_out.shape[1]} raw features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_raw_features(self, x):
        if not self.raw_features_:
            self.logger.debug("No raw features added.")
            dummy_y = pd.Series(np.zeros(len(x)), index=x.index, dtype=np.int32)
            return pd.DataFrame(index=self._pool_target(dummy_y).index)

        self.logger.info("Transforming raw features.")

        # Pooling
        pool_info = [tuple(c.split("__pool=")) for c in self.raw_features_]
        pool_info = pd.DataFrame(pool_info).groupby(0).agg(list)[1].to_dict()
        x_pooled = self._pool_features(x, pool_info)  # type: ignore

        assert set(self.raw_features_) == set(
            x_pooled
        ), "Expected raw features do not match with actual."

        return x_pooled

    @property
    def wav_features_(self):
        out = [str(c) for c in self.features_ if re.search(".+__wav__.+", c)]
        return sorted(out)

    def _fit_transform_wav_features(self, x, y_pooled, update_baseline=True):
        if not self.fit_wavelets:
            self.logger.info("Skipped fitting wavelet-transformed features.")
            dummy_y = pd.Series(np.zeros(len(x)), index=x.index, dtype=np.int32)
            return pd.DataFrame(index=self._pool_target(dummy_y).index)

        self.logger.info("Fitting wavelet-transformed features.")
        self.fit_wavelets = cast(list[str], self.fit_wavelets)  # type hint

        # Pre-initialization
        rng = np.random.default_rng(39478)
        fs = 1.0  # correct sampling frequency only matters for plotting

        # Get (local) peak frequencies of power spectral density
        self.logger.debug("Searching (local) peak frequencies of PSD...")
        peak_freqs = {}
        for col in x:
            freqs, pxx = signal.welch(x=x[col], fs=fs)
            peak_idx, _ = signal.find_peaks(np.log(pxx), prominence=0.3, distance=10)
            peak_freqs[col] = freqs[peak_idx]

        # Initialize column-wavelet iterator
        col_wav_iterator: list[tuple[str, str]] = [
            (col, wav)
            for col in x
            for wav in self.fit_wavelets
            if peak_freqs[col].size > 0
        ]
        if self.strategy in ("random", "smart"):
            # Shuffle iterator
            rng.shuffle(col_wav_iterator)  # type: ignore

        # Initialize watchers (for smart timeout)
        col_watcher = ScoreWatcher(x.keys().to_list())
        wav_watcher = ScoreWatcher(self.fit_wavelets)
        pool_watcher = ScoreWatcher(self.pooling)

        # Initialize data and score tracker
        not_skipped_counter = 0
        all_feats_pooled = []
        all_scores = []

        self.logger.debug("Starting search for wavelet features...")
        start_time = time.time()

        for counter, (col, wav) in enumerate(col_wav_iterator):
            # Check timeout criterion
            if (
                self.strategy in ("random", "smart")
                and time.time() - start_time > self.timeout
            ):
                self.logger.info(
                    f"TIMEOUT: Stopped search for wavelet-transformed features. "
                    f"Examined {counter}/{len(col_wav_iterator)} wavelet-column "
                    f"combinations ({100 * counter / len(col_wav_iterator):.2f}%) "
                    f"whereas {counter - not_skipped_counter} unpromising were skipped."
                )
                break

            # Check column and wavelet watchers to decide whether to continue or skip
            col_counter, col_score = col_watcher[col]
            wav_counter, wav_score = wav_watcher[wav]
            col_score: np.ndarray
            wav_score: np.ndarray
            if self.strategy == "smart" and (
                (
                    col_counter > 10
                    and (col_score < col_watcher.mean() - col_watcher.std()).all()
                )
                or (
                    wav_counter > 10
                    and all(wav_score < wav_watcher.mean() - wav_watcher.std())
                )
            ):
                self.logger.debug(f"SKIPPED: wav `{wav}`, col `{col}`")
                continue

            self.logger.debug(f"Fitting: wav `{wav}`, col `{col}`")

            # Update counter
            not_skipped_counter += 1

            # Decide which pooling to apply
            pooling_instructions = [i for i in self.pooling]  # copy!
            if self.strategy == "random":
                rng.shuffle(pooling_instructions)
                select_n = len(pooling_instructions) // 3  # 33%
                pooling_instructions = pooling_instructions[:select_n]
            elif self.strategy == "smart":
                pooling_instructions = []
                for pool in self.pooling:
                    pool_counter, pool_score = pool_watcher[pool]
                    pool_score: np.ndarray
                    if pool_counter > 10 and all(
                        pool_score < pool_watcher.mean() - pool_watcher.std()
                    ):
                        continue
                    pooling_instructions.append(pool)

            # Use the fact: scale = s2f_const / frequency
            s2f_const = pywt.scale2frequency(wav, scale=1) * fs
            scales = np.round(s2f_const / peak_freqs[col], 2)

            # Extract features, pool and score
            feats = (
                x[col]
                .groupby(level=0, sort=False, group_keys=False)
                .apply(_extract_wavelets, scales=scales, wavelet=wav, name=col)
            )
            feats_pooled = self._pool_features(feats, pooling_instructions, True)
            scores = feats_pooled.apply(self._calc_feature_scores, y=y_pooled, axis=0)  # type: ignore

            # Update watchers
            col_watcher.update(col, scores.sum(1), scores.shape[1])
            wav_watcher.update(wav, scores.sum(1), scores.shape[1])
            for pool in pooling_instructions:
                pool_score = scores.filter(regex=f".*pool={pool}$")
                pool_watcher.update(pool, pool_score.sum(1), pool_score.shape[1])
                # A pooled column might be dropped, e.g. when it contains only NaNs.
                # However, we still want to account for it
                pool_watcher.update(pool, 0.0, feats.shape[1] - pool_score.shape[1])

            # Append good scored features
            good_scores = self.select_scores(
                scores,
                best_n_per_class=20,
                # Update baseline every 20th iteration
                update_baseline=update_baseline and not_skipped_counter % 20 == 0,
            )
            all_scores.append(good_scores)
            all_feats_pooled.append(feats_pooled[good_scores.columns])

            # Debug logging
            if not_skipped_counter % 25 == 0:
                self.logger.debug(
                    f"Finished {not_skipped_counter:5.0f}/{len(col_wav_iterator):5.0f} "
                    f"column-wavelet combinations with a mean time of "
                    f"{(time.time() - start_time) / not_skipped_counter:5.2f} seconds."
                )

        # Finish fitting: concatenate all accepted, pooled features
        x_out = pd.concat(all_feats_pooled, axis=1)
        x_out_scores = pd.concat(all_scores, axis=1)
        if update_baseline:
            self._update_feature_baseline_scores(x_out_scores)

        self.logger.info(f"Accepted {x_out.shape[1]} wavelet-transformed features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_wav_features(self, x):
        if not self.wav_features_:
            self.logger.debug("No wavelet-transformed features added.")
            dummy_y = pd.Series(np.zeros(len(x)), index=x.index, dtype=np.int32)
            return pd.DataFrame(index=self._pool_target(dummy_y).index)

        self.logger.info("Transforming wavelet-transformed features.")

        # Handle wavelet-transform features
        feat_info = []
        for f in self.wav_features_:
            col_name, intermediate = f.split("__wav__")
            wavelet, scale, _ = intermediate.split("__")
            feat_info.append((wavelet, col_name, scale))

        # Group the features by wavelets (moved to index)
        cols = ["wavelet", "col_name", "scales"]
        feat_by_wt = pd.DataFrame(feat_info, columns=cols).groupby("wavelet").agg(list)

        # Extract wavelets
        x_out = []
        for wv, info in feat_by_wt.iterrows():
            columns = sorted(set(info["col_name"]))
            scales = sorted(map(float, set(info["scales"])))

            # Transform and add to list
            for col in columns:
                x_out += [
                    x[col]
                    .groupby(level=0, sort=False, group_keys=False)
                    .apply(_extract_wavelets, scales=scales, wavelet=wv, name=col)
                ]
        x_out = pd.concat(x_out, axis=1)

        # Pooling
        pool_info = [tuple(c.split("__pool=")) for c in self.wav_features_]
        pool_info = pd.DataFrame(pool_info).groupby(0).agg(list)[1].to_dict()
        x_pooled = self._pool_features(x_out, pool_info)  # type: ignore

        assert set(self.wav_features_) == set(
            x_pooled
        ), "Expected wavelet-transform features do not match with actual."

        return x_pooled

    # ----------------------------------------------------------------------
    # Utils

    def _assert_multiindex(
        self, data: PandasType, raise_on_single=True, copy=True
    ) -> tuple[PandasType, bool]:
        data = data.copy() if copy else data
        n_index_cols = len(data.index.names)

        if n_index_cols == 1 and raise_on_single:
            raise ValueError("Data must be multiindexed.")
        elif n_index_cols == 1:
            data.index = pd.MultiIndex.from_product([[0], data.index])
            data_is_multiindexed = False
        elif n_index_cols == 2:
            data_is_multiindexed = True
        else:
            raise ValueError("Data is neither single- nor properly multiindexed.")

        return data, data_is_multiindexed

    def _set_window_size(self, index: pd.Index) -> None:

        if isinstance(self.window_size_, int):
            return

        # Count log sizes
        counts = pd.Series(index=index, dtype=int).fillna(0).groupby(level=0).count()

        if self.mode == "classification":
            # We'll make the window size such that on average there's 5 samples
            # NOTE: ** IMPORTANT ** Window size CANNOT be small, it significantly slows down the window calculations.
            ws = int(min(counts.min(), counts.mean() // 5))

        elif self.mode == "regression":
            ws = 1

        else:
            raise AttributeError(f"Invalid mode '{self.mode}'.")

        # Ensure that window size is an integer and at least 1
        self.window_size_ = max(1, int(ws))

        self.logger.debug(f"Set window size to {self.window_size_}.")

    def _fit_data_to_window_size(
        self, *data: PandasType
    ) -> PandasType | list[PandasType]:
        """
        Fit the data to a multiple of the window size.

        Make sure to always call this function before pooling the data.
        Also, notice that this function uses ``_add_or_remove_tail`` internally.

        Parameters
        ----------
        data : tuple of PandasType
            Data to be fitted to window size.

        Returns
        -------
        data : PandasType or List of PandasType
        """
        self.logger.debug("Fitting data to window size")

        if len(data) < 1:
            raise ValueError("Got no data.")

        data_out = []
        for datum in data:
            # Check datum
            if len(datum.index.names) != 2:
                raise ValueError("Index is not a MultiIndex of size 2.")
            if pd.unique(datum.index.names).size == 1:
                warn("Index names are not unique. Setting them to ['log', 'index'].")
                datum.index.names = ["log", "index"]

            # Add or remove tail
            datum = datum.groupby(level=0, group_keys=False, sort=False).apply(  # type: ignore
                self._add_or_remove_tail
            )
            data_out.append(datum)

        if len(data_out) == 1:
            return data_out[0]
        return data_out

    def _add_or_remove_tail(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Fill or cut the tail to fit the data length to a multiple of the window size.

        This is a helper function to be used with ``_fit_data_to_window_size`` and
        treats the data as being single indexed.

        Parameters
        ----------
        data : pd.DataFrame
            Data to fill, cut or leave its tail.

        Returns
        -------
        pd.DataFrame
            Parsed data.
        """
        self.window_size_ = cast(int, self.window_size_)  # type hint

        if self.window_size_ == 1:
            return data

        tail = data.shape[0] % self.window_size_
        n_missing_in_tail = self.window_size_ - tail
        if 0 < n_missing_in_tail < self.window_size_ / 2:
            # Fill up tail
            add_to_tail = data.iloc[-n_missing_in_tail:]
            data = pd.concat([data, add_to_tail])
        elif tail != 0:
            # Cut tail
            data = data.iloc[:-tail]

        return data

    def _pool_features(
        self,
        data: pd.DataFrame,
        instruction: dict[str, str | list[str] | None] | list[str] | None = None,
        drop_nan_columns=False,
    ) -> pd.DataFrame:
        """
        Pools data with given window size.

        In order to avoid potential problems, make sure that the `data` parameter was
        handled with ``_fit_data_to_window_size`` first.

        Parameters
        ----------
        data : pd.DataFrame
            Data to be pooled.
        instruction : dict of {str: str or list of str or None} or list of str, optional
            Instructions for pooling. Each key corresponds to the column name and
            its value defines the pooling names for the column.
            Default: Will calculate all implemented pools.
        drop_nan_columns : bool
            If false, all columns--no matter how many NaN values they have--will
            be returned. If true, columns with more than 10% of NaN entries will
            be removed.
            You are strongly encouraged to set this parameter only true when you
            are fitting data.

        Returns
        -------
        pd.DataFrame
            Pooled data.
        """
        if not instruction and self.mode == "regression":
            # Use only mean feature (since window_size=1).
            instruction = {str(col): "mean" for col in data}
        elif not instruction:
            instruction = {str(col): None for col in data}  # set default
        elif isinstance(instruction, dict):
            for col, instr in instruction.items():
                instruction[col] = list(get_pool_functions(instr))  # validate
        elif isinstance(instruction, list):
            instr = list(get_pool_functions(instruction))  # validate
            instruction = {str(col): instr for col in data}
        else:
            raise ValueError(f"Invalid instructions: {instruction}")

        # Pooling
        self.logger.debug("Pooling data")
        pooled_data = []
        for col, instr in tqdm(instruction.items(), disable=self.verbose <= 1):
            agg_func = get_pool_functions(instr)
            # Apply
            # Note: For many windows, this becomes slow (up to 2.5s per column
            pooled_col = pl_pool(data[col], self.window_size_, agg_func)  # type: ignore
            pooled_data += [pooled_col]
        pooled_data = pd.concat(pooled_data, axis=1)

        # Sanitize
        if drop_nan_columns:
            rm_mask = pooled_data.isna().sum() > len(pooled_data) / 10
            pooled_data = pooled_data.loc[:, ~rm_mask]

        return sanitize_dataframe(pooled_data)

    def _pool_target(self, target: pd.Series):
        """
        Pools target data with given window size.

        Parameters
        ----------
        target : pd.Series
            Target data to be pooled.

        Returns
        -------
        pd.Series
            Pooled target data.
        """
        assert isinstance(target, pd.Series), "Data must be pandas.Series."
        assert isinstance(self.window_size_, int), "Invalid window size."

        dtype = target.dtype

        if self.mode == "classification":
            agg_func = get_pool_functions("prominent_class")
        elif self.mode == "regression":
            agg_func = get_pool_functions("mean")
        else:
            msg = f"Feature processor has an invalid mode: {self.mode}"
            raise AttributeError(msg)

        out = pl_pool(target, self.window_size_, agg_func)
        out = out.iloc[:, 0].astype(dtype).rename(target.name)

        return out



#  Copyright (c) 2022 by Amplo.

from .feature_processor import FeatureProcessor
from .nop_feature_extractor import NopFeatureExtractor
from .static_feature_extractor import StaticFeatureExtractor
from .temporal_feature_extractor import TemporalFeatureExtractor



#  Copyright (c) 2022 by Amplo.

import json
import warnings

import numpy as np
import pytest
import scipy.stats

from amplo.automl import DriftDetector
from amplo.automl.drift_detection import DataDriftWarning
from amplo.utils.testing import DummyDataSampler, make_num_data


class DummyPredictor(DummyDataSampler):
    def predict(self, data):
        return self.sample_data(len(data))


class TestDriftDetector:
    def test_distribution_fits(self):
        # Setup
        ref, cols = make_num_data(500)
        test = ref.iloc[np.random.permutation(len(ref))[:10]]
        drift = DriftDetector(target="target")
        drift.fit(ref)

        # Checks
        assert len(drift.check(test)) == 0, "Test data found inconsistent"
        assert len(drift.check(ref.max() + 1)) == len(
            ref.columns
        ), "Maxima not detected"
        assert len(drift.check(ref.min() - 1)) == len(
            ref.columns
        ), "Minima not detected"

    def test_storable(self):
        df, cols = make_num_data(100)
        drift = DriftDetector(target="target")
        drift.fit(df)
        json.dumps(drift.bins_)
        pred = np.random.randint(0, 2, (100,))
        old = drift.add_output_bins((), pred)
        drift.add_output_bins(old, pred)

    def test_drift_warning(self):
        """
        Ensure that minor changes in data do not trigger warnings.
        """
        # Create dummy data --
        data_1, _ = make_num_data(500, num_dists=["uniform", "norm"])
        data_2, _ = make_num_data(10, num_dists=["uniform", "norm"])
        data_2["num_0"] += 15
        data_2["num_1"] -= 10
        # NOTE: categorical data is not compared

        # Create dummy predictors
        dummy_model_1 = DummyPredictor()
        dummy_model_2 = DummyPredictor(scipy.stats.randint(0, 10))  # type: ignore

        # Instantiate and fit drift detector
        drift = DriftDetector(target="target", n_bins=10)
        drift.fit(data_1)
        drift.fit_output(dummy_model_1, data_1)  # type: ignore

        # Assert that no DataDriftWarning occurs when given same data and model
        with warnings.catch_warnings(record=True) as caught_warnings:
            # Check drift on input
            drift.check(data_1)
            # Check drift on output
            drift.check_output(dummy_model_1, data_1)  # type: ignore
        if any(
            issubclass(warning.category, DataDriftWarning)
            for warning in caught_warnings
        ):
            raise AssertionError("Unnecessary DataDriftWarning detected")

        # Assert that DataDriftWarning occurs when given new data
        with pytest.warns(DataDriftWarning):
            drift.check(data_2)

        # Assert that DataDriftWarning occurs when given new model
        with pytest.warns(DataDriftWarning):
            drift.check_output(dummy_model_2, data_2)  # type: ignore



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import json
import shutil
from copy import deepcopy
from pathlib import Path
from typing import cast
from warnings import warn

from azure.core.exceptions import ResourceNotFoundError
from requests import HTTPError

from amplo import Pipeline
from amplo.api.databricks import DatabricksJobsAPI
from amplo.api.platform import AmploPlatformAPI
from amplo.api.storage import AzureBlobDataAPI
from amplo.utils import check_dtypes
from amplo.utils.io import merge_logs

__all__ = ["DEFAULT_PIPE_KWARGS", "train_locally", "train_on_cloud"]

DEFAULT_PIPE_KWARGS = {
    "interval_analyse": False,
    "standardize": False,
    "missing_values": "zero",
    "grid_search_timeout": 7200,
    "n_grid_searches": 1,
    "verbose": 1,
}


def _set_default_pipe_kwargs(
    team: str,
    machine: str,
    service: str,
    issue: str,
    model_version: int,
    unhandled_pipe_kwargs: dict | None = None,
) -> dict:
    """
    Inserts default pipeline arguments if not set already.

    Parameters
    ----------
    team : str
        Name of the team.
    machine : str
        Name of the machine.
    service : str
        Name of the service (a.k.a. category).
    issue : str
        Name of the issue (a.k.a. model).
    model_version : int
        Model version.
    unhandled_pipe_kwargs : dict
        Unhandled keyword arguments for pipeline.

    Returns
    -------
    dict
        Pipeline keyword arguments with imputed values.
    """

    if unhandled_pipe_kwargs is None:
        unhandled_pipe_kwargs = {}

    check_dtypes(
        ("team", team, str),
        ("machine", machine, str),
        ("service", service, str),
        ("issue", issue, str),
        ("model_version", model_version, int),
        ("pipe_kwargs", unhandled_pipe_kwargs, dict),
    )

    pipe_kwargs = deepcopy(DEFAULT_PIPE_KWARGS)

    # Insert unhandled pipeline keyword arguments
    pipe_kwargs.update(unhandled_pipe_kwargs)

    # Set params
    pipe_kwargs["name"] = f"{team} - {machine} - {service} - {issue}"
    pipe_kwargs["target"] = issue
    pipe_kwargs["version"] = model_version

    return pipe_kwargs


def _get_file_delta(
    metadata: dict[str, dict],
    team: str,
    machine: str,
    service: str,
    issue: str,
    version: int,
    azure: tuple[str, str] | bool,
    platform: tuple[str, str] | bool,
) -> dict[str, list[str]]:

    # Initialize
    platform_api = AmploPlatformAPI.from_os_env(
        *(platform if not isinstance(platform, bool) else tuple())
    )
    blob_api = AzureBlobDataAPI.from_os_env(
        *(azure if not isinstance(azure, bool) else tuple())
    )

    # Get file_metadata of previous training
    try:
        trainings = platform_api.list_trainings(
            team, machine, service, issue, version - 1
        )
    except HTTPError:  # group matching query does not exist
        trainings = []

    if trainings:
        settings_path = (
            f"{team}/{machine}/models/{service}/{issue}/"
            + trainings[0].get("version", "-")
            + "/Settings.json"
        )
        try:
            settings = blob_api.read_json(settings_path)
            settings = cast(dict, settings)  # type hint
        except ResourceNotFoundError:
            settings = {}
    else:
        settings = {}

    prev_metadata = settings.get("file_metadata", {})

    # Compare files from previous to current version
    curr_files = [meta["full_path"] for meta in metadata.values()]
    prev_files = [meta["full_path"] for meta in prev_metadata.values()]

    return {
        "new_files": sorted(set(curr_files) - set(prev_files)),
        "removed_files": sorted(set(prev_files) - set(curr_files)),
    }


def train_locally(
    data_dir: str | Path,
    target_dir: str | Path,
    team: str,
    machine: str,
    service: str,
    issue: str,
    pipe_kwargs: dict | None = None,
    model_version: int = 1,
    *,
    healthy_data_dir: str | Path | bool = True,
    working_dir: str | Path = "./tmp",
    azure: tuple[str, str] | bool = False,
    platform: tuple[str, str] | bool | None = None,
) -> bool:
    """
    Locally train a model with given parameters.

    Parameters
    ----------
    data_dir : str or Path
        Directory where data is stored. Note that it must contain subdirectories which
        names depict the issues (e.g. pipe error).
    target_dir : str or Path
        Directory where the trained model files will be copied to.
    team : str
        Name of the team.
    machine : str
        Name of the machine.
    service : str
        Name of the service (a.k.a. category).
    issue : str
        Name of the issue (a.k.a. model).
    pipe_kwargs : dict, optional, default: None
        Keyword arguments for pipeline. Note that defaults will be set.
    model_version : int, default: 1
        Model version.
    healthy_data_dir : str or Path or bool, default: True
        Directory where healthy data is stored.
        If False, no healthy data will be used.
        If True, data will be assumed to be in "../Healthy/Healthy" of `data_dir`.
        Otherwise, will try to read from given path.
    working_dir : str or Path, default: "./tmp"
        Directory where temporary training files will be stored.
        Note that this directory will be deleted again.
    azure : (str, str) or bool, default: False
        Use this parameter to indicate that data is in Azure blob storage.
        If False, it is assumed that data origins from local directory.
        If True, the AzureBlobDataAPI is initialized with default OS env variables.
        Otherwise, it will use the tuple to initialize the api.
    platform : (str, str) or bool or None, default: None
        Use this parameter for selecting data according to Amplo's datalogs.
        If None, its value is set to bool(azure).
        If False, no AmploPlatformAPI will be initialized.
        If True, the AmploPlatformAPI is initialized with default OS env variables.
        Otherwise, it will use the tuple to initialize the api.
    logging : bool, default: True
        Whether to show logging info. Currently only applies to `merge_logs`.

    Returns
    -------
    True
    """

    # Input checks
    pipe_kwargs = _set_default_pipe_kwargs(
        team, machine, service, issue, model_version, pipe_kwargs
    )
    check_dtypes(
        ("data_dir", data_dir, (str, Path)),
        ("target_dir", target_dir, (str, Path)),
        ("working_dir", working_dir, (str, Path)),
    )

    # --- Data ---

    # Read data
    target: str = pipe_kwargs["target"]
    more_folders: list[str | Path]
    if not healthy_data_dir:
        more_folders = []
    elif isinstance(healthy_data_dir, (str, Path)):
        more_folders = [healthy_data_dir]
    else:
        more_folders = [Path(data_dir).parent / "Healthy/Healthy"]
    data, file_metadata = merge_logs(
        data_dir, target, more_folders=more_folders, azure=azure, platform=platform
    )

    # --- Training ---

    # Create temporary working directory
    working_dir = (
        Path(working_dir) / f"{team}_{machine}_{service}_{issue}_{model_version}"
    )
    working_dir.mkdir(parents=True, exist_ok=False)

    # Set up pipeline
    pipeline = Pipeline(main_dir=f"{working_dir.as_posix()}/", **pipe_kwargs)

    # Mirror azure parameter when platform is not set
    if platform is None:
        platform = bool(azure)
    # Indicate new files compared to previous model version
    if azure and platform:
        pipeline.settings["file_delta"] = _get_file_delta(
            file_metadata, team, machine, service, issue, model_version, azure, platform
        )

    # Train
    pipeline.fit(data, metadata=file_metadata)

    # --- Post training ---

    # Move training files into target directory
    shutil.move(working_dir / "Production" / f"v{pipeline.version}", target_dir)

    # Delete temporary working directory
    shutil.rmtree(working_dir)

    return True


def train_on_cloud(
    job_id: int,
    model_id: int,
    team: str,
    machine: str,
    service: str,
    issue: str,
    pipe_kwargs: dict | None = None,
    model_version: int = 1,
    *,
    train_id: int | None = None,
    host_os: str | None = None,
    access_token_os: str | None = None,
) -> dict[str, int]:
    """
    Train a model with given parameters on the cloud (Databricks).

    Notes
    -----
    Make sure to have set the following environment variables:
        - ``DATABRICKS_INSTANCE``
        (see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication).
        - ``DATABRICKS_ACCESS_TOKEN`` (see Databricks > User Settings > Access tokens).

    Note two important differences to ``DatabricksJobsAPI.run_job``.
    The "pipe_kwargs" key of ``notebook_params``:
        - will be JSON dumped to a string for you.
        - gets default values imputed if not given.

    Parameters
    ----------
    job_id : int
        Job ID in Databricks.
    model_id : int
        Model training ID in Amplo's platform.
    team : str
        Name of the team.
    machine : str
        Name of the machine.
    service : str
        Name of the service (a.k.a. category).
    issue : str
        Name of the issue (a.k.a. model).
    pipe_kwargs : dict, optional, default: None
        Keyword arguments for pipeline. Note that defaults will be set.
    model_version : int, default: 1
        Model version.
    train_id : int, optional for now
        Training ID, used for error handling. Will be a required parameter in future editions.
    host_os : str, optional, default: None
        Key in the os environment for the Databricks host.
    access_token_os : str, optional, default: None
        Key in the os environment for the Databricks access token.

    Returns
    -------
    dict of {str: int}
        If response is success (200), ``run_id`` (globally unique key of newly triggered
        run) is one of the present keys.
    """
    if not train_id:
        warn("train_id will be a required parameter in the future.", DeprecationWarning)

    # Input checks
    pipe_kwargs = _set_default_pipe_kwargs(
        team, machine, service, issue, model_version, pipe_kwargs
    )
    check_dtypes(
        ("job_id", job_id, int),
        ("model_id", model_id, int),
        ("team", team, str),
        ("machine", machine, str),
        ("service", service, str),
        ("issue", issue, str),
        ("train_id", train_id, int),
        ("host_os", host_os, (type(None), str)),
        ("access_token_os", access_token_os, (type(None), str)),
    )

    # Set up notebook params
    notebook_params = {
        "team": team,
        "machine": machine,
        "service": service,
        "issue": issue,
        "model_id": model_id,
        "train_id": train_id,
        "pipe_kwargs": json.dumps(pipe_kwargs),
    }

    # Send request
    api = DatabricksJobsAPI.from_os_env(host_os, access_token_os)
    return api.run_job(job_id=job_id, notebook_params=notebook_params)



#  Copyright (c) 2022 by Amplo.

import os

import numpy as np
import pandas as pd
import pytest

from amplo import Pipeline
from tests import create_test_folders, get_all_modeller_models, rmtree


class TestPipeline:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_main_predictors(self, mode, make_data):
        # Test mode
        data = make_data
        pipeline = Pipeline(n_grid_searches=0, plot_eda=False, extract_features=False)
        pipeline.fit(data)
        x_c = pipeline.transform(data)

        models = get_all_modeller_models(mode)
        for model in models:
            model.fit(x_c, data["target"])
            pipeline.best_model_ = model
            pipeline.predict(data)
            assert isinstance(
                pipeline.main_predictors_, dict
            ), f"Main predictors not dictionary: {type(pipeline.main_predictors_)}"

    @pytest.mark.parametrize("mode", ["classification"])
    def test_no_dirs(self, mode, make_data):
        data = make_data
        pipeline = Pipeline(no_dirs=True, n_grid_searches=0, extract_features=False)
        pipeline.fit(data)
        assert not os.path.exists("Auto_ML"), "Directory created"

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_mode_detector(self, mode, make_data):
        data = make_data
        pipeline = Pipeline()
        pipeline._mode_detector(data)
        assert pipeline.mode == mode

    @pytest.mark.parametrize("mode", ["classification"])
    def test_create_folders(self, mode, make_data):
        data = make_data
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(data)

        # Test Directories
        assert os.path.exists("Auto_ML")
        assert os.path.exists("Auto_ML/Model.joblib")
        assert os.path.exists("Auto_ML/Settings.json")

    @pytest.mark.parametrize("mode", ["classification"])
    def test_capital_target(self, mode, make_data):
        data = make_data
        data["TARGET"] = data["target"]
        data = data.drop("target", axis=1)
        pipeline = Pipeline(target="TARGET", n_grid_searches=0, extract_features=False)
        pipeline.fit(data)

    def test_read_dir(self):
        create_test_folders("data", n_samples=100, n_features=5)

        # First multi-class
        df1 = Pipeline(target="Class_1")._read_data("data")
        assert "class_1" in df1
        assert len(df1.keys()) == 6
        assert isinstance(df1.index, pd.MultiIndex)
        assert df1.index.get_level_values(0).dtype == "object"

        # Second binary class
        df2 = Pipeline()._read_data("data", "Class_1")
        assert (df1.index == df2.index).all()
        assert "class_1" in df2
        assert set(df2["class_1"].values) == {0, 1}

        # Trigger errors
        with pytest.raises(ValueError):
            Pipeline()._read_data("data", "NoClass")
        with pytest.raises(ValueError):
            Pipeline()._read_data("data")
        with pytest.raises(ValueError):
            Pipeline()._read_data("NoFolder")

        # Cleanup
        rmtree("data")

    def test_read_numpy(self):
        x = np.random.normal(0, 1, (100, 10))
        y = np.random.normal(0, 1, 100)

        # Normal
        data = Pipeline(target="tArGeT")._read_data(x, y)
        assert "target" in data
        assert len(data.keys()) == 11

        # With series
        data = Pipeline()._read_data(x, pd.Series(y, name="label"))
        assert "label" in data
        assert len(data.keys()) == 11

        # Trigger errors
        with pytest.raises(NotImplementedError):
            Pipeline()._read_data(x, None)
        with pytest.raises(ValueError):
            Pipeline()._read_data(x, np.random.normal(0, 1, 101))

    def test_read_pandas(self):
        index = pd.Index(np.linspace(101, 200, 100))
        x = pd.DataFrame(
            {
                "x1": np.random.normal(0, 1, 100),
                "x2": np.random.normal(0, 1, 100),
            },
            index=index,
        )
        y = pd.Series(np.random.normal(0, 1, 100), name="target", index=index)
        data = x.copy()
        data["tArGeT"] = y

        # Normal
        d1 = Pipeline()._read_data(x, y)
        d2 = Pipeline()._read_data(data, "tArGeT")
        d3 = Pipeline(target="tArGeT")._read_data(data)
        assert d1.equals(d2) and d1.equals(d3)
        assert "target" in d1
        assert len(d1.keys()) == 3

        # Trigger errors
        with pytest.raises(ValueError):
            # y too long
            Pipeline()._read_data(x, pd.Series(np.random.normal(0, 1, 101)))
        with pytest.warns(Warning):
            # different index
            Pipeline()._read_data(x, y.reset_index(drop=True))
        with pytest.warns(Warning):
            # different name
            Pipeline()._read_data(
                x, pd.Series(np.random.normal(0, 1, 100), index=index, name="label2")
            )
        with pytest.raises(NotImplementedError):
            # wrong type
            Pipeline()._read_data(x, {"a": 1})  # type: ignore
        with pytest.raises(ValueError):
            # Missing target
            Pipeline()._read_data(data, "label")
        with pytest.raises(ValueError):
            data2 = data.rename(columns={"target": "target2"})
            Pipeline()._read_data(data2)



#  Copyright (c) 2022 by Amplo.

"""
Defines various pooling functions.
"""

from __future__ import annotations

from typing import Callable

import numpy as np
import polars as pl
from numpy.typing import ArrayLike
from polars import internals as pli

# ----------------------------------------------------------------------
# Basics


def root_mean_square(column: str | pli.Series) -> pli.Expr | float:
    if isinstance(column, str):
        column = pli.col(column)
    return (column**2).mean() ** 0.5


def sum_values(column: str | pli.Series) -> pli.Expr | float:
    if isinstance(column, str):
        column = pli.col(column)
    return column.sum()


# ----------------------------------------------------------------------
# Characteristics


def abs_energy(column: str | pli.Series) -> pli.Expr | float:
    if isinstance(column, str):
        column = pli.col(column)
    return column.dot(column)


def abs_max(column: str | pli.Series) -> pli.Expr | float:
    if isinstance(column, pli.Series):
        return column.abs().max()
    return pli.col(column).abs().max()


def n_mean_crossings(column: str | pli.Series) -> pli.Expr | float:
    """
    Calculates the number of crossings of x on mean.

    A crossing is defined as two sequential values where the first value is lower than
    mean and the next is greater, or vice-versa.
    """
    if not isinstance(column, pli.Series):
        column = pli.col(column)
    column = (column - column.mean()).sign().diff()
    return (column.fill_nan(0).fill_null(0) != 0).sum()


# ----------------------------------------------------------------------
# Difference


def abs_sum_of_changes(column: str | pli.Series) -> pli.Expr | float:
    if not isinstance(column, pli.Series):
        column = pli.col(column)
    return column.diff().abs().sum()


def mean_of_changes(column: str | pli.Series) -> pli.Expr | float:
    if not isinstance(column, pli.Series):
        column = pli.col(column)
    return column.diff().mean()


def abs_mean_of_changes(column: str | pli.Series) -> pli.Expr | float:
    if not isinstance(column, pli.Series):
        column = pli.col(column)
    return column.diff().abs().mean()


def cid_ce(column: str | pli.Series, normalize: bool = True) -> float:
    """
    Calculates an estimate for a time series complexity.

    Parameters
    ----------
    column
    normalize

    Returns
    -------

    """
    if not isinstance(column, pli.Series):
        column = pli.col(column)
    if normalize:
        std = column.std()
        if std != 0:
            column = (column - column.mean()) / std
        else:
            return 0.0
    column = column.diff()
    return np.sqrt(column.dot(column))


# ----------------------------------------------------------------------
# Special


def prominent_class(x: ArrayLike):
    return np.bincount(x, minlength=2).argmax()


# ----------------------------------------------------------------------
# Globals

POOL_FUNCTIONS = {
    # --- Basics ---
    "min": pl.min,
    "max": pl.max,
    "mean": pl.mean,
    "std": pl.std,
    "median": pl.median,
    "variance": pl.var,
    "kurtosis": pl.Expr.kurtosis,
    "skew": pl.Expr.skew,
    "root_mean_square": root_mean_square,
    "sum_values": sum_values,
    # --- Characteristics ---
    "entropy": pl.Expr.entropy,
    "abs_energy": abs_energy,
    "abs_max": abs_max,
    # TODO: Add linear trend feature for `slope` and `stderror`. We have to
    #  adjust our pooling function to support target data as an input.
    #  c.f. tsfresh.feature_extraction.linear_trend
    # "linear_trend_slope": ...,
    # "linear_trend_stderror": ...,
    "n_mean_crossings": n_mean_crossings,
    # --- Difference ---
    "abs_sum_of_changes": abs_sum_of_changes,
    "mean_of_changes": mean_of_changes,
    "abs_mean_of_changes": abs_mean_of_changes,
    "cid_ce": cid_ce,
}
MORE_POOL_FUNCTIONS = {
    "prominent_class": prominent_class,
}


def get_pool_functions(requests: str | list[str] | None = None) -> dict[str, Callable]:
    """
    Get pool functions.

    Parameters
    ----------
    requests : str or List of str, optional
        If None, defaults (POOL_FUNCTIONS) are returned.

    Returns
    -------
    dict of {str: typing.Callable}
    """
    if requests is None:
        return POOL_FUNCTIONS
    elif isinstance(requests, str):
        requests = [requests]

    pools = {}
    for req in requests:
        if req in POOL_FUNCTIONS:
            pools[req] = POOL_FUNCTIONS[req]
        elif req in MORE_POOL_FUNCTIONS:
            pools[req] = MORE_POOL_FUNCTIONS[req]
        else:
            raise ValueError(f"Invalid pool function request: {req}")

    return pools



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

import time
from datetime import datetime
from typing import TypeVar

import numpy as np
import pandas as pd
from sklearn import ensemble, linear_model, model_selection, svm

from amplo.base import LoggingMixin
from amplo.classification import CatBoostClassifier, LGBMClassifier, XGBClassifier
from amplo.regression import CatBoostRegressor, LGBMRegressor, XGBRegressor
from amplo.utils.logging import get_root_logger

__all__ = ["ClassificationType", "Modeller", "ModelType", "RegressionType"]


logger = get_root_logger().getChild("Modeller")


ClassificationType = TypeVar(
    "ClassificationType",
    CatBoostClassifier,
    ensemble.BaggingClassifier,
    linear_model.RidgeClassifier,
    linear_model.LogisticRegression,
    LGBMClassifier,
    svm.SVC,
    XGBClassifier,
)
RegressionType = TypeVar(
    "RegressionType",
    CatBoostRegressor,
    ensemble.BaggingRegressor,
    linear_model.LinearRegression,
    LGBMRegressor,
    svm.SVR,
    XGBRegressor,
)
ModelType = TypeVar(
    "ModelType",
    CatBoostClassifier,
    CatBoostRegressor,
    ensemble.BaggingClassifier,
    ensemble.BaggingRegressor,
    linear_model.LinearRegression,
    linear_model.LogisticRegression,
    linear_model.RidgeClassifier,
    LGBMClassifier,
    LGBMRegressor,
    svm.SVC,
    svm.SVR,
    XGBClassifier,
    XGBRegressor,
)


class Modeller(LoggingMixin):
    """
    Modeller for classification or regression tasks.

    Supported models:
        - linear models from ``scikit-learn``
        - random forest from ``scikit-learn``
        - bagging model from ``scikit-learn``
        - ~~gradient boosting from ``scikit-learn``~~
        - ~~histogram-based gradient boosting from ``scikit-learn``~~
        - XGBoost from DMLC
        - Catboost from Yandex
        - LightGBM from Microsoft

    Parameters
    ----------
    mode : str
        Model mode. Either `regression` or `classification`.
    shuffle : bool
        Whether to shuffle samples from training / validation.
    n_splits : int
        Number of cross-validation splits.
    objective : str
        Performance metric to optimize. Must be a valid string for
        `sklearn.metrics.get_scorer`.
    samples : int
        Hypothetical number of samples in dataset. Useful to manipulate behavior
        of `return_models()` function.
    needs_proba : bool
        Whether the modelling needs a probability.

    See Also
    --------
    [Sklearn scorers](https://scikit-learn.org/stable/modules/model_evaluation.html
    """

    def __init__(
        self,
        target: str | None = None,
        mode: str = "classification",
        cv: model_selection.BaseCrossValidator | None = None,
        objective: str | None = None,
        samples: int | None = None,
        needs_proba: bool = True,
        verbose: int = 1,
    ):
        super().__init__(verbose=verbose)
        if mode not in ("classification", "regression"):
            raise ValueError(f"Unsupported mode: {mode}")

        # Parameters
        self.target = target
        self.cv = cv
        self.objective = objective
        self.mode = mode
        self.samples = samples
        self.needs_proba = needs_proba
        self.results = pd.DataFrame(
            columns=[
                "date",
                "model",
                "params",
                "score",
                "worst_case",
                "time",
            ]
        )

        # Update CV if not provided
        if self.cv is None:
            if self.mode == "classification":
                self.cv = model_selection.StratifiedKFold(n_splits=3)
            elif self.mode == "regression":
                self.cv = model_selection.KFold(n_splits=3)

    def fit(self, data: pd.DataFrame) -> pd.DataFrame:
        if not self.target:
            raise ValueError("Can only fit when target is provided.")
        if self.target not in data:
            raise ValueError(f"Target column not in dataframe: {self.target}")

        self.samples = len(data)
        y = data[self.target]
        x = data.drop(self.target, axis=1)

        # Convert to NumPy
        x = np.array(x)
        y = np.array(y).ravel()

        # Models
        self.models = self.return_models()

        # Loop through models
        for model in self.models:

            # Time & loops through Cross-Validation
            t_start = time.time()
            scores = model_selection.cross_val_score(
                model, x, y, scoring=self.objective
            )
            score = sum(scores) / len(scores)
            run_time = time.time() - t_start

            # Append results
            result = {
                "date": datetime.today().strftime("%d %b %y"),
                "model": type(model).__name__,
                "params": model.get_params(),
                "score": score,
                "worst_case": np.mean(scores) - np.std(scores),
                "time": run_time,
            }
            self.results = pd.concat(
                [self.results, pd.Series(result).to_frame().T], ignore_index=True
            )
            self.logger.info(
                f"{result['model'].ljust(30)} {self.objective}: "
                f"{result['worst_case']:15.4f}    training time:"
                f" {result['time']:.1f} s"
            )

        # Return results
        return self.results

    def return_models(self):
        """
        Get all models that are considered appropriate for training.

        Returns
        -------
        list of ModelType
            Models that apply for given dataset size and mode.
        """
        models = []

        # All classifiers
        if self.mode == "classification":
            # The thorough ones
            if not self.samples or self.samples < 25000:
                models.append(svm.SVC(kernel="rbf", probability=self.needs_proba))
                models.append(ensemble.BaggingClassifier())
                # models.append(ensemble.GradientBoostingClassifier()) == XG Boost
                models.append(XGBClassifier())

            # The efficient ones
            else:
                # models.append(ensemble.HistGradientBoostingClassifier()) == LGBM
                models.append(LGBMClassifier())

            # And the multifaceted ones
            if not self.needs_proba:
                models.append(linear_model.RidgeClassifier())
            else:
                models.append(linear_model.LogisticRegression())
            models.append(CatBoostClassifier())
            models.append(ensemble.RandomForestClassifier())

        elif self.mode == "regression":
            # The thorough ones
            if not self.samples or self.samples < 25000:
                models.append(svm.SVR(kernel="rbf"))
                models.append(ensemble.BaggingRegressor())
                # models.append(ensemble.GradientBoostingRegressor()) == XG Boost
                models.append(XGBRegressor())

            # The efficient ones
            else:
                # models.append(ensemble.HistGradientBoostingRegressor()) == LGBM
                models.append(LGBMRegressor())

            # And the multifaceted ones
            models.append(linear_model.LinearRegression())
            models.append(CatBoostRegressor())
            models.append(ensemble.RandomForestRegressor())

        return models



#  Copyright (c) 2022 by Amplo.

import warnings
from typing import Any
from unittest.mock import DEFAULT

import pytest

from amplo import Pipeline
from amplo.training import DEFAULT_PIPE_KWARGS, _set_default_pipe_kwargs


def test_default_pipe_kwargs():
    with warnings.catch_warnings(record=True) as warns:
        Pipeline(**DEFAULT_PIPE_KWARGS)
    assert not any(isinstance(w.message, UserWarning) for w in warns), warns


def test_set_default_pipe_kwargs():
    params: dict[str, Any] = dict(team="1", machine="2", service="3", issue="4")
    params["model_version"] = 1

    # Case 1: no kwargs are given
    pipe_kwargs = _set_default_pipe_kwargs(**params, unhandled_pipe_kwargs=None)
    # Test that default values are set
    assert {key: pipe_kwargs[key] for key in DEFAULT_PIPE_KWARGS} == DEFAULT_PIPE_KWARGS
    # Test that name, target and version are set
    assert pipe_kwargs["name"] == "1 - 2 - 3 - 4"
    assert pipe_kwargs["target"] == params["issue"]
    assert pipe_kwargs["version"] == params["model_version"]

    # Case 2: kwargs are given and shall not be overwritten
    pipe_kwargs = _set_default_pipe_kwargs(
        **params, unhandled_pipe_kwargs={"issue": "labels", "verbose": 2}
    )
    # Test that given values came through
    assert pipe_kwargs["target"] == params["issue"]
    assert pipe_kwargs["verbose"] == 2
    # Test that the other default parameters are not overwritten
    default_pipe_kwargs = DEFAULT_PIPE_KWARGS
    default_pipe_kwargs.pop("verbose")
    assert {key: pipe_kwargs[key] for key in default_pipe_kwargs} == DEFAULT_PIPE_KWARGS


def test_train_on_premise():
    pytest.skip("Not yet implemented.")


def test_train_on_cloud():
    pytest.skip("Not yet implemented.")



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

from warnings import warn

import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import LabelEncoder

from amplo.base import LoggingMixin
from amplo.utils.util import check_dtypes, clean_column_names, clean_feature_name

__all__ = ["clean_feature_name", "DataProcessor"]


class DataProcessor(LoggingMixin):
    def __init__(
        self,
        target: str | None = None,
        include_output: bool = True,
        drop_datetime: bool = False,
        drop_constants: bool = False,
        drop_duplicate_rows: bool = False,
        missing_values: str = "interpolate",
        outlier_removal: str = "clip",
        z_score_threshold: int = 4,
        version: int = 1,
        verbose: int = 1,
    ):
        """
        Preprocessing Class. Cleans a dataset into a workable format.
        Deals with Outliers, Missing Values, duplicate rows, data types (floats,
        categorical and dates), Not a Numbers, Infinities.

        Parameters
        ----------
        target : str
            Column name of target variable
        include_output : bool
            Whether to include output in the data
        drop_datetime : bool
            Whether to drop datetime columns
        drop_contstants : bool
            If False, does not remove constants
        drop_duplicate_rows : bool
            If False, does not remove constant columns
        missing_values : {"remove_rows", "remove_cols", "interpolate", "mean", "zero"}
            How to deal with missing values.
        outlier_removal : {"quantiles", "z-score", "clip", "none"}
            How to deal with outliers.
        z_score_threshold : int
            If outlier_removal="z-score", the threshold is adaptable
        version : int
            Versioning the output files
        verbose : int
            How much to print
        """
        super().__init__(verbose=verbose)
        # Type checks
        check_dtypes(
            ("target", target, (type(None), str)),
            # ignore "int_cols" and "float_cols" as they're deprecated and unused
            ("include_output", include_output, bool),
            ("drop_datetime", drop_datetime, bool),
            ("drop_constants", drop_constants, bool),
            ("drop_duplicate_rows", drop_duplicate_rows, bool),
            ("z_score_threshold", z_score_threshold, int),
            ("drop_duplicate_rows", drop_duplicate_rows, bool),
            ("version", version, int),
            ("verbose", verbose, int),
        )

        # Integrity checks
        mis_values_algo = ["remove_rows", "remove_cols", "interpolate", "mean", "zero"]
        if missing_values not in mis_values_algo:
            raise ValueError(
                f"Missing values algorithm not implemented, pick from {mis_values_algo}"
            )
        out_rem_algo = ["quantiles", "z-score", "clip", "none"]
        if outlier_removal not in out_rem_algo:
            raise ValueError(
                f"Outlier Removal algorithm not implemented, pick from {out_rem_algo}"
            )

        # Arguments
        self.version = version
        self.include_output = include_output
        self.drop_datetime = drop_datetime
        self.target = target

        # Algorithms
        self.missing_values = missing_values
        self.outlier_removal = outlier_removal
        self.z_score_threshold = z_score_threshold
        self.drop_constants = drop_constants
        self.drop_duplicate_rows = drop_duplicate_rows

        # Fitted Settings
        self.num_cols_ = []
        self.bool_cols_ = []
        self.cat_cols_ = []
        self.date_cols_ = []
        self.dummies_ = {}
        self.q1_ = None
        self.q3_ = None
        self.means_ = None
        self.stds_ = None
        self.label_encodings_ = []

        # Info for Documenting
        self.is_fitted_ = False
        self.removed_duplicate_rows_ = 0
        self.removed_duplicate_columns_ = 0
        self.removed_outliers_ = 0
        self.imputed_missing_values_ = 0
        self.removed_constant_columns_ = 0

    def fit_transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Fits this data cleaning module and returns the transformed data.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # Clean column names and apply renaming
        data, self.rename_dict_ = clean_column_names(data)
        if self.target is not None:
            self.target = self.rename_dict_.get(self.target, None)

        # Remove target
        if not self.include_output and self.target is not None and self.target in data:
            data = data.drop(self.target, axis=1)

        # Remove Duplicates
        data = self.remove_duplicates(data, rows=self.drop_duplicate_rows)

        # Infer data-types
        data = self.infer_data_types(data)

        # Convert data types
        data = self.convert_data_types(data, fit_categorical=True)

        # Remove outliers
        data = self.fit_remove_outliers(data)

        # Remove missing values
        data = self.remove_missing_values(data)

        # Remove Constants
        if self.drop_constants:
            data = self.remove_constants(data)

        # Encode or decode target
        data = self.encode_labels(data, fit=True)

        # Finish
        self.is_fitted_ = True

        return data

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Function that takes existing settings (including dummies), and transforms new
        data.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        if not self.is_fitted_:
            raise ValueError("Transform only available for fitted objects.")

        # Clean column names and apply renaming
        data = data.rename(columns=self.rename_dict_)

        # Impute columns
        data = self._impute_columns(data)

        # Convert data types
        data = self.convert_data_types(data, fit_categorical=False)

        # Remove outliers
        data = self.remove_outliers(data)

        # Remove duplicates
        data = self.remove_duplicates(data)

        # Remove missing values
        data = self.remove_missing_values(data)

        # Encode or drop target
        data = self.encode_labels(data, fit=False)

        return data

    def get_settings(self) -> dict:
        """
        Get settings to recreate fitted object.
        """
        assert self.is_fitted_, "Object not yet fitted."
        settings = {
            "include_output": self.include_output,
            "drop_datetime": self.drop_datetime,
            "drop_constants": self.drop_constants,
            "drop_duplicate_rows": self.drop_duplicate_rows,
            "missing_values": self.missing_values,
            "outlier_removal": self.outlier_removal,
            "z_score_threshold": self.z_score_threshold,
            "rename_dict_": self.rename_dict_,
            "num_cols_": self.num_cols_,
            "bool_cols_": self.bool_cols_,
            "date_cols_": self.date_cols_,
            "cat_cols_": self.cat_cols_,
            "label_encodings_": self.label_encodings_,
            "means_": (
                self.means_.to_json() if isinstance(self.means_, pd.Series) else None
            ),
            "stds_": (
                self.stds_.to_json() if isinstance(self.stds_, pd.Series) else None
            ),
            "q1_": self.q1_.to_json() if isinstance(self.q1_, pd.Series) else None,
            "q3_": self.q3_.to_json() if isinstance(self.q3_, pd.Series) else None,
            "dummies_": self.dummies_,
            "imputed_missing_values_": self.imputed_missing_values_,
            "removed_outliers_": self.removed_outliers_,
            "removed_constant_columns_": self.removed_constant_columns_,
            "removed_duplicate_rows_": self.removed_duplicate_rows_,
            "removed_duplicate_columns_": self.removed_duplicate_columns_,
        }
        return settings

    def load_settings(self, settings: dict):
        """
        Loads settings from dictionary and recreates a fitted object
        """
        self.rename_dict_ = settings.get("rename_dict_", {})
        self.num_cols_ = settings.get("num_cols_", settings.get("num_cols", []))
        self.bool_cols_ = settings.get("bool_cols_", settings.get("bool_cols", []))
        self.date_cols_ = settings.get("date_cols_", settings.get("date_cols", []))
        self.cat_cols_ = settings.get("cat_cols_", settings.get("cat_cols", []))
        self.label_encodings_ = settings.get("label_encodings_", [])
        self.include_output = settings.get("include_output", True)
        self.drop_datetime = settings.get("drop_datetime", False)
        self.drop_constants = settings.get("drop_constants", False)
        self.drop_duplicate_rows = settings.get("drop_duplicate_rows", False)
        self.missing_values = settings.get("missing_values", [])
        self.outlier_removal = settings.get("outlier_removal", [])
        self.z_score_threshold = settings.get("z_score_threshold", [])
        self.means_ = settings.get("means_", settings.get("_means", None))
        self.stds_ = settings.get("stds_", settings.get("_stds", None))
        self.q1_ = settings.get("q1_", settings.get("_q1", None))
        self.q3_ = settings.get("q3_", settings.get("_q3", None))
        for key in ["means_", "stds_", "q1_", "q3_"]:
            if getattr(self, key):
                setattr(self, key, pd.read_json(getattr(self, key), typ="series"))
        self.dummies_ = settings.get("dummies_", settings.get("dummies", {}))
        self.is_fitted_ = True
        return self

    def infer_data_types(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        In case no data types are provided, this function infers the most likely data
        types

        parameters
        ----------
        data : pd.DataFrame

        returns
        -------
        data : pd.DataFrame
        """
        # Iterate through keys
        for key in data.keys():

            # Skip target -- we don't want to convert it
            if key == self.target:
                continue

            # Remove NaN for feature identification
            f = data[key]
            f = f[~f.isna()].infer_objects()

            # Integer and Float
            if pd.api.types.is_integer_dtype(f) or pd.api.types.is_float_dtype(f):
                self.num_cols_.append(key)
                continue

            # Datetime
            elif pd.api.types.is_datetime64_any_dtype(f):
                self.date_cols_.append(key)
                continue

            # Booleans
            elif pd.api.types.is_bool_dtype(f):
                self.bool_cols_.append(key)
                continue

            # Strings / Objects
            elif pd.api.types.is_object_dtype(f):

                # Check numerical
                numeric = pd.to_numeric(f, errors="coerce", downcast="integer")
                if numeric.isna().sum() < 0.3 * len(f):
                    self.num_cols_.append(key)

                    # Update data and continue
                    data[key] = numeric
                    continue

                # Check date
                date = pd.to_datetime(
                    f.astype("str"),
                    errors="coerce",
                    infer_datetime_format=True,
                )
                if date.isna().sum() < 0.3 * len(f):
                    self.date_cols_.append(key)
                    continue

                # Check categorical variable
                if data[key].nunique() < max(10, len(data) // 4):
                    self.cat_cols_.append(key)
                    continue

            # Else not found
            warn(f"Couldn't identify feature: {key}")
        return data

    def convert_data_types(
        self, data: pd.DataFrame, fit_categorical: bool = True
    ) -> pd.DataFrame:
        """
        Cleans up the data types of all columns.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        fit_categorical : bool
            Whether to fit the categorical encoder

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # Drop unused columns & datetime columns
        data = data.drop(
            [
                k
                for k in data.keys()
                if k
                not in self.num_cols_
                + self.date_cols_
                + self.bool_cols_
                + self.cat_cols_
                + [self.target]
            ],
            axis=1,
        )

        if self.date_cols_ and self.drop_datetime:
            warn(
                f"Data contains datetime columns but are removed: '{self.date_cols_}'",
                UserWarning,
            )
            data = data.drop(self.date_cols_, axis=1)

        # Or convert to datetime (before done only on subset)
        elif self.date_cols_:
            for key in self.date_cols_:
                data[key] = pd.to_datetime(
                    data[key], errors="coerce", infer_datetime_format=True
                )

        # Integer columns
        for key in self.bool_cols_:
            data[key] = data[key].fillna(False).astype(np.int64)

        # Float columns
        for key in self.num_cols_:
            data[key] = pd.to_numeric(data[key], errors="coerce", downcast="float")

        # Categorical columns
        if fit_categorical:
            data = self._fit_cat_cols(data)
        else:
            assert self.is_fitted_, (
                ".convert_data_types() was called with fit_categorical=False, while "
                "categorical encoder is not yet fitted."
            )
        data = self._transform_cat_cols(data)

        return data

    def _fit_cat_cols(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Encoding categorical variables always needs a scheme. This fits the scheme.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        for key in self.cat_cols_:
            # Clean the categorical variables
            data[key] = data[key].apply(clean_feature_name)
            is_nan = data[key].isna().any()

            # Get dummies, convert & store
            dummies = pd.get_dummies(data[key], prefix=key, dummy_na=is_nan)
            self.dummies_[key] = dummies.keys().tolist()

        return data

    def _transform_cat_cols(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Converts categorical variables according to fitted scheme.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        for key in self.cat_cols_:
            # Clean column
            data[key] = data[key].apply(clean_feature_name)

            # Transform
            value = self.dummies_[key]
            dummies = [i[len(key) + 1 :] for i in value]
            str_values = data[key].astype("str").values
            data[value] = np.equal.outer(str_values, dummies).astype(np.int64)
            data = data.drop(key, axis=1)
        return data

    def remove_duplicates(self, data: pd.DataFrame, rows: bool = False) -> pd.DataFrame:
        """
        Removes duplicate columns and rows.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        rows : bool
            Whether to remove duplicate rows --> not desirable to maintain timelines

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # Note down
        n_rows, n_columns = len(data), len(data.keys())

        # Remove Duplicates
        if rows:
            data = data.drop_duplicates()
        data = data.loc[:, ~data.columns.duplicated()]  # type: ignore

        # Note
        self.removed_duplicate_columns = n_columns - len(data.keys())
        self.removed_duplicate_rows = n_rows - len(data)

        return data

    def remove_constants(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Removes constant columns

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        columns = len(data.keys())

        # Remove Constants
        const_cols = [
            col for col in data if data[col].nunique() == 1 and col != self.target
        ]
        data = data.drop(columns=const_cols)

        # Note
        self.removed_constant_columns = columns - len(data.keys())

        return data

    def fit_remove_outliers(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Fits parameters necessary to remove outliers and removes them using remove_outliers method.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # With quantiles
        if self.outlier_removal == "quantiles":
            self.q1_ = data[self.num_cols_].quantile(0.25)
            self.q3_ = data[self.num_cols_].quantile(0.75)

        # By z-score
        elif self.outlier_removal == "z-score":
            self.means_ = data[self.num_cols_].mean(skipna=True, numeric_only=True)
            self.stds_ = data[self.num_cols_].std(skipna=True, numeric_only=True)
            self.stds_[self.stds_ == 0] = 1

        return self.remove_outliers(data)

    def remove_outliers(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Removes outliers

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # With Quantiles
        if self.outlier_removal == "quantiles":
            self.removed_outliers = (
                (data[self.num_cols_] > self.q3_).sum().sum()
                + (data[self.num_cols_] < self.q1_).sum().sum()
            ).tolist()
            data[self.num_cols_] = data[self.num_cols_].mask(
                data[self.num_cols_] < self.q1_
            )
            data[self.num_cols_] = data[self.num_cols_].mask(
                data[self.num_cols_] > self.q3_
            )

        # With z-score
        elif self.outlier_removal == "z-score":
            z_score = abs((data[self.num_cols_] - self.means_) / self.stds_)
            self.removed_outliers = (
                (z_score > self.z_score_threshold).sum().sum().tolist()  # type: ignore
            )
            data[self.num_cols_] = data[self.num_cols_].mask(
                z_score > self.z_score_threshold  # type: ignore
            )

        # With clipping
        elif self.outlier_removal == "clip":
            self.removed_outliers = (
                (data[self.num_cols_] > 1e12).sum().sum()
                + (data[self.num_cols_] < -1e12).sum().sum()
            ).tolist()
            data[self.num_cols_] = data[self.num_cols_].clip(lower=-1e12, upper=1e12)
        return data

    def remove_missing_values(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Fills missing values (infinities and "not a number"s)

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # Replace infinities
        data = data.replace([np.inf, -np.inf], np.nan)

        # Note
        self.imputed_missing_values = data[self.num_cols_].isna().sum().sum().tolist()

        # Removes all rows with missing values
        if self.missing_values == "remove_rows":
            data.dropna(axis=0, inplace=True)

        # Removes all columns with missing values
        elif self.missing_values == "remove_cols":
            data.dropna(axis=1, inplace=True)

        # Fills all missing values with zero
        elif self.missing_values == "zero":
            data = data.fillna(0)

        # Mean and Interpolate require more than 1 value, use zero if less
        elif self.missing_values in ("interpolate", "mean") and len(data) <= 1:
            data = data.fillna(0)

        # Linearly interpolates missing values
        elif self.missing_values == "interpolate":
            # Get all non-date_cols & interpolate
            non_date = np.setdiff1d(data.keys().to_list(), self.date_cols_)
            data[non_date] = data[non_date].interpolate(limit_direction="both")

            # Fill rest (date & more missing values cols)
            data = self._interpolate_dates(data)

        # Fill missing values with column mean
        elif self.missing_values == "mean":
            non_date = np.setdiff1d(data.keys().to_list(), self.date_cols_)
            data[non_date] = data[non_date].fillna(data.mean())

            # Fill dates
            data = self._interpolate_dates(data)

        return data

    def _interpolate_dates(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Unfortunately pandas does not support this out of the box. PR was made, but
        closed pre-merged. https://github.com/pandas-dev/pandas/pull/21915

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        for key in self.date_cols_:
            if data[key].isna().any():
                unix = data[key].astype("int64")
                unix[unix < 0] = np.nan  # NaT are -9e10
                unix = unix.interpolate(method="bfill").interpolate("pad")
                data[key] = pd.to_datetime(unix, unit="ns")
        return data

    def encode_labels(self, data: pd.DataFrame, fit: bool) -> pd.DataFrame:
        """En- or decodes target column of `data`

        Parameters
        ----------
        data : pd.DataFrame
            input data
        fit : bool
            Whether to (re)fit the label encoder

        Returns
        -------
        data : pd.DataFrame
            With the encoded labels
        """
        # Get labels and encode / decode
        if not self.target or self.target not in data:
            return data
        if not self.include_output:
            return data.drop(self.target, axis=1)

        # Split output
        labels = data[self.target]

        # Check whether it's classification
        if labels.dtype == object or labels.nunique() <= labels.size / 2:

            # Create encoder
            encoder = LabelEncoder()
            if fit is True:
                encoder.fit(labels)
                self.label_encodings_ = pd.Series(encoder.classes_).to_list()
            elif not self.label_encodings_:
                raise NotFittedError("Encoder it not yet fitted")
            else:
                encoder.fit(self.label_encodings_)

            # Encode
            data[self.target] = encoder.transform(labels)
            return data

        # It's probably a regression task, thus no encoding needed
        warn(UserWarning("Labels are probably for regression. No encoding happened..."))
        return data

    def decode_labels(self, data: np.ndarray) -> pd.Series:
        """Decode labels from numerical dtype to original value

        Parameters
        ----------
        data : np.ndarray
            Input data

        Returns
        -------
        data : pd.Series
            With labels encoded

        Raises
        ------
        NotFittedError
            When `except_not_fitted` is True and label encoder is not fitted
        """
        # Checks
        if len(self.label_encodings_) == 0:
            raise NotFittedError(
                "Encoder it not yet fitted. Try first calling `encode_target` "
                "to set an encoding"
            )

        # Create encoder
        encoder = LabelEncoder()
        encoder.classes_ = np.array(self.label_encodings_)

        # Decode
        return pd.Series(encoder.inverse_transform(data))  # type: ignore

    def _impute_columns(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        *** For production ***
        If a dataset is missing certain columns, this function looks at all registered
        columns and fills them with
        zeros.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        data : pd.DataFrame
        """
        # Impute
        keys = self.date_cols_ + self.num_cols_ + self.bool_cols_ + self.cat_cols_
        to_impute = [k for k in keys if k not in data]
        data[to_impute] = np.zeros((len(data), len(to_impute)))

        # Warn
        if len(to_impute) > 0:
            warn(f"Imputed {len(to_impute)} missing columns! {to_impute}")
        return data

    def prune_features(self, features: list):
        """
        For use with AutoML.Pipeline. We practically never use all features. Yet this
        processor imputes any missing features. This causes redundant operations,
        memory, and warnings. This function prunes the features to avoid that.

        parameters
        ----------
        features : list
            Required features (NOTE: include required features for extracted)
        """
        hash_features = dict([(k, 0) for k in features])
        self.date_cols_ = [f for f in self.date_cols_ if f in hash_features]
        self.num_cols_ = [f for f in self.num_cols_ if f in hash_features]
        self.bool_cols_ = [f for f in self.bool_cols_ if f in hash_features]
        self.cat_cols_ = [f for f in self.cat_cols_ if f in hash_features]



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

from warnings import warn

import numpy as np
import pandas as pd
from scipy import stats

from amplo.base import BasePredictor, LoggingMixin
from amplo.utils import hist_search


class DataDriftWarning(Warning):
    pass


class DriftDetector(LoggingMixin):
    # todo add second order pdf fit
    # todo add subsequence drift detection

    def __init__(
        self,
        target: str,
        n_bins: int = 500,
        sigma: int = 3,
        with_pdf: bool = False,
        verbose: int = 1,
    ):
        """Detects data drift in streamed input data.

        Supports only  numerical data as off now.
        - Numerical data is check by the discrete distribution
        - Categorical by new columns ()
        - Dates by their recency

        parameters
        ----------
        target : str
        n_bins : int, default =  500
            Number of bins to put the data in
        sigma : int, default = 3
        with_pdf : bool, default = False
        verbose : int, default = 1
        """
        super().__init__(verbose=verbose)
        # Copy kwargs
        self.target = target
        self.n_bins = n_bins
        self.with_pdf = with_pdf
        self.sigma = sigma

        # Initialize
        self.bins_ = {}
        self.output_bins_ = (None, None)
        self.distributions_ = {}
        self.is_fitted_ = False

    def fit(self, data: pd.DataFrame) -> "DriftDetector":
        """
        Fits the class object
        """
        # Numerical
        numeric = data.select_dtypes(include=np.number)  # type: ignore
        self._fit_bins(numeric)
        self._fit_distributions(numeric)

        # Categorical
        # TODO: check for new values

        # Dates
        # TODO: check for recency

        self.is_fitted_ = True
        return self

    def check(self, data: pd.DataFrame | pd.Series):
        """
        Checks a new dataframe for distribution drift.
        """
        violations = []

        # Convert series
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data.to_dict(), index=[0])

        # Numerical
        numeric = data.select_dtypes(include=np.number)  # type: ignore
        violations.extend(self._check_bins(numeric))
        violations.extend(self._check_distributions(numeric))

        return violations

    def fit_output(self, model: BasePredictor, data: pd.DataFrame):
        """
        Additionally to detecting input drift, we should also detect output drift. When
        the distribution of predicted outcomes change, it's often a sign that some under
        laying dynamics are shifting.
        """
        # If it's a classifier and has predict_proba, we use that :)
        if hasattr(model, "predict_proba"):
            prediction = model.predict_proba(data)[:, 1]  # type: ignore
        else:
            prediction = model.predict(data)

        ma, mi = max(prediction), min(prediction)
        y, x = np.histogram(
            prediction,
            bins=self.n_bins,
            range=(mi - (ma - mi) / 10, ma + (ma - mi) / 10),
        )
        self.output_bins_ = (x.tolist(), y.tolist())

    def check_output(self, model: BasePredictor, data: pd.DataFrame, add: bool = False):
        """
        Checks the predictions of a model.
        """
        if not self.is_fitted_:
            raise ValueError("Object not yet fitted.")

        # If it's a classifier and has predict_proba, we use that :)
        if hasattr(model, "predict_proba"):
            prediction = model.predict_proba(data)[:, 1]  # type: ignore
        else:
            prediction = model.predict(data)

        # Unpack predictions
        x, y = self.output_bins_
        assert isinstance(x, list) and isinstance(y, list)

        count_drifts = 0
        for value in prediction:
            ind = hist_search(x, value)
            if ind == -1 or y[ind] <= 0:
                # Drift detected
                count_drifts += 1
        if count_drifts > 0:
            severity = count_drifts / len(prediction) * 100
            warn(f"Output drift detected! Severity: {severity:.2f}%", DataDriftWarning)

        # Add new output
        if add:
            y += np.histogram(prediction, bins=x)
            return y

    def get_weights(self) -> dict:
        """
        Gets the weights of the fitted object.
        Useful to save :)
        """
        return {
            "bins": self.bins_,
            "output_bins": self.output_bins_,
            "distributions": self.distributions_,
        }

    def load_weights(self, weights):
        """
        Sets the weights of the object to recreate a previously fitted object.

        Parameters
        ----------
        weights : typing.Dict[str, dict or tuple]
            Weights of a (fitted) object.
            Expected keys are:
            - "bins" (dict):
                Bins dictionary with bins and quantities for all numeric keys.
            - "output_bins" (tuple):
                Output bins.
            - "distributions" (dict):
                Fitted distributions for all numeric keys.
        """
        self.bins_ = weights.get("bins", {})
        self.output_bins_ = weights.get("output_bins", (None, None))
        self.distributions_ = weights.get("distributions", {})
        return self

    def _fit_bins(self, data: pd.DataFrame):
        """
        Fits a histogram on each numerical column.
        """
        # Fit numerical
        for key in data.keys():
            ma, mi = data[key].max(), data[key].min()
            y, x = np.histogram(
                data[key],
                bins=self.n_bins,
                range=(mi - (ma - mi) / 10, ma + (ma - mi) / 10),
            )
            self.bins_[key] = (x.tolist(), y.tolist())

    def _check_bins(self, data: pd.DataFrame, add: bool = False):
        """
        Checks if the current data falls into bins
        """
        violations = []

        for key in data.keys():
            # Get bins
            if key not in self.bins_:
                violations.append(key)
                continue
            x, y = self.bins_[key]

            # Check bins
            if isinstance(data, pd.DataFrame):
                for value in data[key].values:
                    ind = hist_search(x, value)
                    if ind == -1 or y[ind] <= 0:
                        violations.append(key)
                        break
            elif isinstance(data, pd.Series):
                ind = hist_search(x, data[key])
                if ind == -1 or (
                    y[ind] <= 0
                    and y[max(0, ind - 1)] <= 0
                    and y[min(self.n_bins, ind + 1)] <= 0
                ):
                    violations.append(key)

            # Add data
            if add:
                y += np.histogram(data[key], bins=x)
                self.bins_[key] = (x, y)

        if len(violations) > 0:
            warn(
                DataDriftWarning(
                    f"Drift detected! "
                    f"{len(violations)} features outside training bins: {violations}"
                )
            )

        return violations

    def _fit_distributions(self, data: pd.DataFrame):
        """
        Fits a distribution on each numerical column.
        """
        if self.with_pdf:
            distributions = ["gamma", "beta", "dweibull", "dgamma"]
            distances = []
            fitted = []

            # Iterate through numerical columns
            for key in data.keys():
                y, x = np.histogram(data[key], normed=True)  # type: ignore
                x = (x + np.roll(x, -1))[:-1] / 2.0  # Get bin means

                # Iterate through distributions
                for distribution in distributions:
                    # Fit & Get PDF
                    dist = getattr(stats, distribution)

                    # Multiple order fit
                    params = dist.fit(data[key])
                    fitted_pdf = dist.pdf(
                        x, loc=params[-2], scale=params[-1], *params[:-2]
                    )

                    # Analyse
                    distances.append(sum((y - fitted_pdf) ** 2))
                    fitted.append(
                        {
                            "distribution": distribution,
                            "params": params,
                        }
                    )

                # Select lowest
                self.distributions_[key] = fitted[np.argmin(distances)]

    def _check_distributions(self, data: pd.DataFrame) -> list:
        """
        Checks whether the new data falls within the fitted distributions
        """
        # Init
        violations = []

        if self.with_pdf:
            # Check all numerical columns
            for key in data.keys():
                dist = getattr(stats, self.distributions_[key]["distribution"])
                params = self.distributions_[key]["params"]
                probabilities = dist.pdf(
                    data[key].values, loc=params[-2], scale=params[-1], *params[:-2]
                )

                if any(p < self.sigma for p in probabilities):
                    violations.append(key)
                    continue

            if len(violations) > 0:
                warn(
                    f"Drift detected! {len(violations)} features outside training bins:"
                    f" {violations}",
                    DataDriftWarning,
                )

        return violations

    def add_output_bins(self, old_bins: tuple, prediction: np.ndarray | pd.Series):
        """
        Just a utility, adds new data to an old distribution.
        """
        if len(old_bins) != 0:
            x, y = old_bins
            yn = np.histogram(prediction, bins=x)[0].tolist()
            y = [y[i] + yn[i] for i in range(len(y))]
        else:
            y, x = np.histogram(prediction, bins=self.n_bins)
            x, y = x.tolist(), y.tolist()
        return x, y



#  Copyright (c) 2022 by Amplo.

import pytest

from amplo import Pipeline
from amplo.observation._base import ProductionWarning
from tests import DelayedRandomPredictor


class TestProductionObserver:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_prediction_latency(self, mode, make_x_y):
        pytest.skip()



#  Copyright (c) 2022 by Amplo.

import json

import numpy as np
import pandas as pd
import pytest

from amplo.automl import DataProcessor
from amplo.utils.data import check_dataframe_quality


@pytest.mark.usefixtures("make_mode")
class TestMode:
    target: str
    data: pd.DataFrame

    def test_mode(self):
        dp = DataProcessor(self.target)
        cleaned = dp.fit_transform(self.data)
        assert check_dataframe_quality(cleaned)


class TestDataProcessor:
    def test_interpolation(self):
        dp = DataProcessor()
        assert dp.missing_values == "interpolate", "Unexpected default value"
        data = pd.DataFrame({"a": [1, np.nan, np.nan, np.nan, 5], "b": [1, 2, 3, 4, 5]})
        cleaned = dp.fit_transform(data)
        assert cleaned["a"].astype(int).tolist() == [1, 2, 3, 4, 5]

    def test_type_detector_no_nan(self):
        dp = DataProcessor()
        data = pd.DataFrame(
            {
                "cat": ["a", "b", "c", "b"],
                "int": [1, 2, 3, 4],
                "date": [
                    "2020-01-01",
                    "2020-01-03",
                    "2020-01-04",
                    "2020-01-05",
                ],
                "bool": [True, False, True, False],
                "float": [0.1, 0.2, 0.3, 0.4],
            }
        )
        cleaned = dp.fit_transform(data)

        assert {
            "cat_a",
            "cat_b",
            "cat_c",
            "int",
            "date",
            "bool",
            "float",
        } == set(cleaned.columns), "Unexpected columns"
        assert len(dp.num_cols_) == 2
        assert set(dp.num_cols_) == {"float", "int"}
        assert len(dp.bool_cols_) == 1
        assert set(dp.bool_cols_) == {"bool"}
        assert len(dp.cat_cols_) == 1
        assert set(dp.cat_cols_) == {"cat"}

    def test_type_detector_with_nan(self):
        dp = DataProcessor()
        data = pd.DataFrame(
            {
                "cat": ["a", "b", np.nan, "c", "b"],
                "int": [1, 2, 3, 4, np.nan],
                "date": [
                    "2020-01-01",
                    np.nan,
                    "2020-01-03",
                    "2020-01-04",
                    "2020-01-05",
                ],
                "bool": [True, False, np.nan, True, False],
                "float": [0.1, 0.2, 0.3, 0.4, np.nan],
            }
        )
        cleaned = dp.fit_transform(data)

        assert {
            "cat_a",
            "cat_b",
            "cat_nan",
            "cat_c",
            "int",
            "date",
            "bool",
            "float",
        } == set(cleaned.columns), "Unexpected columns"
        assert len(dp.num_cols_) == 2
        assert set(dp.num_cols_) == {"float", "int"}
        assert len(dp.bool_cols_) == 1
        assert set(dp.bool_cols_) == {"bool"}
        assert len(dp.cat_cols_) == 1
        assert set(dp.cat_cols_) == {"cat"}

    def test_missing_values(self):
        data = pd.DataFrame(
            {
                "a": ["a", "b", np.nan, "c", "b"],
                "b": [1, 2, 3, 4, np.nan],
                "c": ["2020-01-01", np.nan, "2020-01-03", "2020-01-04", "2020-01-05"],
                "d": [1, 2, 3, 4, 5],
            }
        )

        # Remove rows
        # Despite having NaNs in 3 rows, we expect only one row to be removed:
        # - "a" is categorical, thus encoded and has no NaNs anymore
        # So we go from rows 5 -> 3, cols 4 -> 6
        dp = DataProcessor(missing_values="remove_rows")
        cleaned = dp.fit_transform(data)
        assert cleaned.shape == (3, 7), f"Did not remove NaNs as expected \n{cleaned}"
        assert not cleaned.isna().values.any(), "DataFrame still contains NaNs"

        # Remove cols
        # Similar argumentation as above.
        dp = DataProcessor(missing_values="remove_cols")
        cleaned = dp.fit_transform(data)
        assert cleaned.shape == (5, 5), "Did not remove NaNs as expected"
        assert not cleaned.isna().values.any(), "DataFrame still contains NaNs"

        # Replace with 0
        dp = DataProcessor(missing_values="zero")
        cleaned = dp.fit_transform(data)
        assert (cleaned.loc[2, ["a_a", "a_b", "a_c"]] == 0).all()
        assert cleaned["b"][4] == 0
        assert not cleaned.isna().values.any(), "DataFrame still contains NaNs"

        # Interpolate
        dp = DataProcessor(missing_values="interpolate")
        cleaned = dp.fit_transform(data)
        assert not cleaned.isna().values.any(), "DataFrame still contains NaNs"

        # Fill with mean
        dp = DataProcessor(missing_values="mean")
        cleaned = dp.fit_transform(data)
        assert (cleaned.loc[2, ["a_a", "a_b", "a_c"]] == 0).all()
        assert cleaned["b"][4] == 2.5
        assert not cleaned.isna().values.any(), "DataFrame still contains NaNs"

    def test_classification_target(self):
        data = pd.DataFrame(
            {
                "a": [2, 2, 1, 1, 2],
                "b": ["class1", "class2", "class1", "class2", "class1"],
            }
        )

        # Numerical not starting at 0
        dp = DataProcessor(target="a")
        cleaned = dp.fit_transform(data)
        assert set(cleaned["a"]) == {0, 1}

        # Categorical
        dp = DataProcessor(target="b")
        cleaned = dp.fit_transform(data)
        assert set(cleaned["b"]) == {0, 1}

    def test_outliers(self):
        x = pd.DataFrame(
            {
                "a": [*(23 * [1]), 1e15],
                "b": np.linspace(0, 1, 24),
                "target": np.linspace(0, 1, 24).tolist(),
            }
        )

        # Clip
        dp = DataProcessor(outlier_removal="clip", target="target")
        xt = dp.fit_transform(x)
        assert xt.values.max() < 1e15, "Outlier not removed"
        assert not xt.isna().values.any(), "NaN found"

        # z-score
        dp = DataProcessor(outlier_removal="z-score", target="target")
        xt = dp.fit_transform(x)
        assert xt.values.max() < 1e15, "Outlier not removed"
        assert not xt.isna().values.any(), "NaN found"
        assert np.isclose(
            dp.transform(pd.DataFrame({"a": [1e14], "b": [1]})).values.max(), 1e14
        )
        assert dp.transform(pd.DataFrame({"a": [1e16], "b": [1]})).values.max() == 1

        # Quantiles
        dp = DataProcessor(outlier_removal="quantiles", target="target")
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"
        assert dp.transform(pd.DataFrame({"a": [2], "b": [-2]})).values.max() == 0

    def test_duplicates(self):
        # Create a DataFrame that contains two columns with the same name
        x = pd.DataFrame({"a": [1, 2, 1], "a_copy": [1, 2, 1], "b": [3, 1, 3]})
        x = x.rename(columns={"a_copy": "a"})
        dp = DataProcessor(drop_duplicate_rows=True)
        xt = dp.fit_transform(x)
        assert len(xt) == 2, "Didn't remove duplicate rows"
        assert len(xt.keys()) == 2, "Didn't remove duplicate columns"

    def test_constants(self):
        x = pd.DataFrame({"a": [1, 1, 1, 1, 1], "b": [1, 2, 3, 5, 6]})
        dp = DataProcessor(drop_constants=True)
        xt = dp.fit_transform(x)
        assert "a" not in xt.keys(), "Didn't remove constant column"

    def test_dummies(self):
        x = pd.DataFrame({"a": ["a", "b", "c", "b", "c", "a"]})
        dp = DataProcessor()
        xt = dp.fit_transform(x)
        assert "a" not in xt.keys(), "'a' still in keys"
        assert "a_b" in xt.keys(), "a_b missing"
        assert "a_c" in xt.keys(), "a_c missing"
        xt2 = dp.transform(pd.DataFrame({"a": ["a", "c"]}))
        assert np.allclose(
            xt2.values,
            pd.DataFrame({"a_a": [1, 0], "a_b": [0, 0], "a_c": [0, 1]}).values,
        ), "Converted not correct"

    def test_nan_categorical(self):
        # Setup
        dp = DataProcessor()
        data = pd.DataFrame({"a": ["hoi", np.nan, np.nan, np.nan]})
        cleaned = dp.fit_transform(data)

        # Tests
        assert "a" in dp.cat_cols_, "Did not recognize categorical column."
        assert "a_hoi" in cleaned, f"Cat column not properly converted: {list(cleaned)}"

    def test_settings(self):
        # todo add tests for different arguments, was failing for
        #  outlier_removal=z-score.
        target = "target"
        x = pd.DataFrame(
            {
                "a": ["a", "b", "c", "b", "c", "a"],
                "b": [1, 1, 1, 1, 1, 1],
                target: ["a", "b", "c", "b", "c", "a"],
            }
        )
        dp = DataProcessor(target=target, include_output=False, drop_constants=True)
        xt = dp.fit_transform(x)
        assert len(xt.keys()) == x["a"].nunique()
        settings = dp.get_settings()
        dp2 = DataProcessor()
        dp2.load_settings(settings)
        assert isinstance(dp2.get_settings()["label_encodings_"], list)
        xt2 = dp2.transform(pd.DataFrame({"a": ["a", "b"], "b": [1, 2]}))
        assert np.allclose(
            pd.DataFrame(
                {"b": [1.0, 2.0], "a_a": [1, 0], "a_b": [0, 1], "a_c": [0, 0]}
            ).values,
            xt2.values,
        )

    def test_pruner(self):
        x = pd.DataFrame({"a": ["a", "b", "c", "b", "c", "a"], "b": [1, 1, 1, 1, 1, 1]})
        dp = DataProcessor()
        dp.fit_transform(x)
        dp.prune_features(["b"])
        assert dp.num_cols_ == ["b"]
        assert dp.cat_cols_ == []

    def test_json_serializable(self):
        x = pd.DataFrame(
            {"a": ["a", "b", "c", "b", "c", "a"], "b": [1, "missing", 1, 1, 1, 1]}
        )
        for o in ["quantiles", "z-score", "clip", "none"]:
            for mv in ["remove_rows", "remove_cols", "interpolate", "mean", "zero"]:
                dp = DataProcessor(outlier_removal=o, missing_values=mv)
                dp.fit_transform(x)
                json.dumps(dp.get_settings())

    def test_cat_target(self):
        df = pd.DataFrame(
            {"a": ["a", "b", "c", "b", "c", "a"], "b": [1, 2, 3, 4, 5, 6]}
        )
        dp = DataProcessor(target="a")
        xt = dp.fit_transform(df)
        assert "a" in xt
        assert np.allclose(np.array(xt["a"].values), np.array([0, 1, 2, 1, 2, 0]))

    def test_target_encoding(self):
        target = "target"
        y_values = ["a", "b", "c", "b", "c", "a"]
        data = pd.DataFrame({target: y_values})
        dp = DataProcessor(target=target)
        encoded = dp.encode_labels(data, fit=True)
        assert data[target].nunique() == encoded[target].nunique()
        assert encoded[target].min() == 0, "Encoding must start at zero"
        assert pd.api.types.is_integer_dtype(
            encoded[target]
        ), "Encoding must be of dtype `int`"
        decoded = dp.decode_labels(np.array(encoded[target].values))
        assert (
            decoded == y_values
        ).all(), "Decoding does not result in original dataframe"



#  Copyright (c) 2022 by Amplo.

import pandas as pd
import pytest

from amplo.automl import Modeller
from tests import rmtree


@pytest.fixture(scope="class", autouse=True)
def clean():
    yield None
    rmtree("tmp/")


@pytest.mark.usefixtures("make_mode")
class TestModelling:
    folder = "tmp/"
    mode: str
    objective: str
    data: pd.DataFrame

    def test_modeller(self):
        mod = Modeller(target="target", mode=self.mode, objective=self.objective)
        mod.fit(self.data)
        # Tests
        assert isinstance(
            mod.results, pd.DataFrame
        ), "Results should be type pd.DataFrame"
        assert len(mod.results) != 0, "Results empty"
        assert mod.results["score"].max() < 1, "R2 needs to be smaller than 1"
        assert (
            not mod.results["worst_case"].isna().any()
        ), "worst_case shouldn't contain NaN"
        assert not mod.results["time"].isna().any(), "Time shouldn't contain NaN"
        assert "date" in mod.results.keys()
        assert "model" in mod.results.keys()
        assert "params" in mod.results.keys()

    @pytest.mark.parametrize("n_samples", [100, 100_000])
    def test_return(self, n_samples):
        Modeller(mode=self.mode, samples=n_samples).return_models()



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path

import colorlog

__all__ = [
    "add_file_handler",
    "del_file_handlers",
    "get_root_logger",
    "del_root_logger",
]


# ------------------------------------------------------------------------------
# Filters


class TimeFilter(logging.Filter):
    def filter(self, record):
        # Check if previous logged
        if not hasattr(self, "last"):
            self.last = record.relativeCreated

        # Calc & add delta
        delta = datetime.fromtimestamp(
            record.relativeCreated / 1000.0
        ) - datetime.fromtimestamp(self.last / 1000.0)
        record.relative = f"{(delta.seconds + delta.microseconds / 1e6):.2f}"

        # Update last
        self.last = record.relativeCreated

        return True


class NameFilter(logging.Filter):
    """
    Logging filter that ignores child names of loggers that inherit from ``AutoML``.
    """

    def filter(self, record):
        split_name = record.name.split(".", 1)
        if split_name[0] == "AmploML":
            record.name = split_name[0]
        return True


def _add_filters(handler: logging.Handler):
    handler.addFilter(TimeFilter())
    handler.addFilter(NameFilter())


# ------------------------------------------------------------------------------
# Loggers

_ROOT_LOGGER: logging.Logger | None = None


def _create_logger() -> logging.Logger:
    """
    Creates a new logger that also captures warnings from `warnings.warn()`.

    Returns
    -------
    logging.Logger
        New logger instance.
    """

    # Get custom logger
    logger = logging.getLogger("AmploML")
    logger.setLevel("INFO")

    # Set console handler
    console_formatter = colorlog.ColoredFormatter(
        "%(white)s%(asctime)s %(blue)s[%(name)s]%(log_color)s[%(levelname)s] "
        "%(white)s%(message)s %(light_black)s<%(filename)s:%(lineno)d> (%(relative)ss)",
        datefmt="%H:%M",
    )
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.NOTSET)
    console_handler.setFormatter(console_formatter)
    _add_filters(console_handler)
    logger.addHandler(console_handler)

    # Capture warnings from `warnings.warn(...)`
    logging.captureWarnings(True)
    py_warnings_logger = logging.getLogger("py.warnings")
    warnings_formatter = colorlog.ColoredFormatter(
        "%(white)s%(asctime)s %(blue)s[%(name)s]%(log_color)s[%(levelname)s] "
        "%(white)s%(message)s",
        datefmt="%H:%M",
    )
    warnings_handler = logging.StreamHandler()
    warnings_handler.setLevel(logging.WARNING)
    warnings_handler.setFormatter(warnings_formatter)
    warnings_handler.terminator = ""  # suppress unnecessary newline
    _add_filters(warnings_handler)
    py_warnings_logger.addHandler(warnings_handler)

    return logger


def add_file_handler(file_path: str | Path):
    """
    Add a file handler to the root logger.

    Parameters
    ----------
    file_path : str or Path
        Path where the logger should write to.

    Raises
    ------
    AttributeError
        When the root logger is not properly initialized (None).
    """

    global _ROOT_LOGGER

    if not isinstance(_ROOT_LOGGER, logging.Logger):
        raise AttributeError(
            "The root logger is not initialized properly. "
            "Did you call `get_root_logger()`? "
            f"Root logger: {_ROOT_LOGGER}"
        )

    # Set file handler
    file_formatter = logging.Formatter(
        "%(asctime)s [%(name)s][%(levelname)s] %(message)s  "
        "<%(filename)s:%(lineno)d> (%(relative)ss)",
        datefmt="%H:%M",
    )
    file_handler = logging.FileHandler(file_path, mode="a")
    file_handler.setLevel(logging.NOTSET)
    file_handler.setFormatter(file_formatter)
    _add_filters(file_handler)
    _ROOT_LOGGER.addHandler(file_handler)


def del_file_handlers():
    """
    Delete all file handlers in the root logger.

    Raises
    ------
    AttributeError
        When the root logger is not properly initialized (None).
    """
    global _ROOT_LOGGER

    if not isinstance(_ROOT_LOGGER, logging.Logger):
        raise AttributeError(
            "The root logger is not initialized properly. "
            "Did you call `get_root_logger()`? "
            f"Root logger: {_ROOT_LOGGER}"
        )

    for handler in _ROOT_LOGGER.handlers:
        if isinstance(handler, logging.FileHandler):
            _ROOT_LOGGER.removeHandler(handler)


def get_root_logger():
    """
    Get the root logger. If not yet done the logger will be initialized.
    """

    global _ROOT_LOGGER

    # Do not initialize the same logger multiple times
    if isinstance(_ROOT_LOGGER, logging.Logger):
        return _ROOT_LOGGER
    # First time called -> initialize logger
    _ROOT_LOGGER = _create_logger()

    return _ROOT_LOGGER


def del_root_logger():
    """Reset the root logger and set it to None."""

    global _ROOT_LOGGER
    _ROOT_LOGGER = None



#  Copyright (c) 2022 by Amplo.

import numpy as np
from sklearn.ensemble import StackingClassifier as _StackingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from amplo.classification._base import BaseClassifier
from amplo.utils import check_dtypes, get_model


def _get_default_estimators(n_samples=None):
    defaults = {
        "DecisionTreeClassifier": DecisionTreeClassifier,
        "GaussianNB": GaussianNB,
        "KNeighborsClassifier": KNeighborsClassifier,
        "LogisticRegression": LogisticRegression,
    }
    if n_samples is not None and n_samples < 5000:
        defaults.update({"SVC": SVC})

    return defaults


def _make_estimator_stack(estimators, add_defaults=True, n_samples=None):
    """
    Make a stack of estimators for the stacking model.

    Parameters
    ----------
    estimators : list of str
        List of estimators for the stack.
    add_defaults : bool, default: True
        Whether to add default estimators to the stack.
    n_samples : int, optional
        (Expected) number of samples to determine the default estimators.

    Returns
    -------
    list of (str, estimator)
        Stack of estimators.
    """
    check_dtypes(
        ("estimators", estimators, list),
        *[(f"estimators_item: `{est}`", est, str) for est in estimators],
    )

    # Initialize
    stack = {}

    # Add default models
    if add_defaults:
        for model_name, model in _get_default_estimators(n_samples).items():
            stack[model_name] = model()  # initialize model

    # Add Amplo models
    for model_name in estimators:
        if model_name in stack:
            # Skip default models
            continue
        stack[model_name] = get_model(model_name)

    return [(key, value) for key, value in stack.items()]


def _get_final_estimator(n_samples=None, n_features=None):
    check_dtypes(
        ("n_samples", n_samples, (type(None), int)),
        ("n_features", n_features, (type(None), int)),
    )

    many_samples = not n_samples or n_samples > 10_000
    many_features = not n_features or n_features > 100

    solver = "lbfgs" if many_samples or many_features else "sag"
    return LogisticRegression(max_iter=2000, solver=solver)


class StackingClassifier(BaseClassifier):
    """
    Stacking classifier.

    Parameters
    ----------
    add_to_stack : list of str, optional
        List of estimators for the stack of estimators.
    add_defaults_to_stack : bool, default: True
        Whether to add default estimators to the stack. This option will be set to True
        when the `add_to_stack` parameter is None.
    n_samples : int, optional
        (Expected) number of samples.
    n_features : int, optional
        (Expected) number of features.
    verbose : {0, 1, 2}, default: 0
        Verbose logging.
    **model_params : Any
        Model parameters for underlying models.
    """

    _add_to_settings = ["_mean", "_std", *BaseClassifier._add_to_settings]
    model: _StackingClassifier  # type hint

    def __init__(
        self,
        add_to_stack=None,
        add_defaults_to_stack=True,
        n_samples=None,
        n_features=None,
        verbose=0,
        **model_params,
    ):
        check_dtypes(("add_defaults_to_stack", add_defaults_to_stack, bool))

        # Set attributes
        if add_to_stack is None:
            add_to_stack = []
            add_defaults_to_stack = True
        model = _StackingClassifier(
            _make_estimator_stack(add_to_stack, add_defaults_to_stack, n_samples),
            _get_final_estimator(n_samples, n_features),
        )
        model.set_params(**model_params)

        # Set attributes
        self.add_to_stack = add_to_stack
        self.add_defaults_to_stack = add_defaults_to_stack
        self.n_samples = n_samples
        self.n_features = n_features

        super().__init__(model=model, verbose=verbose)

    def _fit(self, x, y=None, **fit_params):
        # When `self.n_samples` or `self.n_features` is None or badly initialized, we
        # reset the stacking estimator as its stack and final estimator depend on that.
        if self.n_samples != x.shape[0] or self.n_features != x.shape[1]:
            self.n_samples, self.n_features = x.shape

            # Get previous model parameters
            prev_model_params = self._get_model_params()

            # Init new stacking classifier
            self.model = _StackingClassifier(
                _make_estimator_stack(
                    self.add_to_stack or [], self.add_defaults_to_stack, self.n_samples
                ),
                _get_final_estimator(self.n_samples, self.n_features),
            )

            # Update model parameters from previous model
            model_params = self._get_model_params()
            for key in set(model_params).intersection(prev_model_params):
                model_params[key] = prev_model_params[key]
            self.model.set_params(**model_params)

        # Normalize
        mean = np.mean(x, axis=0)
        std = np.std(x, axis=0)
        std[std == 0] = 1
        self._mean = np.asarray(mean).reshape(-1).tolist()
        self._std = np.asarray(std).reshape(-1).tolist()
        x -= mean
        x /= std

        # Fit model
        self.model.fit(x, y)

    def _predict(self, x, y=None, **kwargs):
        mean = np.array(self._mean)
        std = np.array(self._std)
        return self.model.predict((x - mean) / std, **kwargs).reshape(-1)

    def predict_proba(self, x, **kwargs):
        self.check_is_fitted()
        mean = np.array(self._mean)
        std = np.array(self._std)
        return self.model.predict_proba((x - mean) / std, **kwargs)

    def _get_model_params(self, deep=True):
        model_params = self.model.get_params(deep)

        non_serializable = ["estimators", "final_estimator"]
        if deep:
            non_serializable.extend(name for name, _ in model_params["estimators"])
        for key in non_serializable:
            model_params.pop(key)

        return model_params



#  Copyright (c) 2022 by Amplo.

import pytest
import requests

from amplo.api.platform import (
    AmploPlatformAPI,
    PlatformSynchronizer,
    _format_version,
    upload_model,
)
from amplo.utils.testing import make_production_data
from tests.unit.api import TestAPI


class TestPlatform(TestAPI):
    def test_upload(self):
        """
        This only tests that upload_latest_model doesn't error.
        - Good to check whether the file exists in the cloud.
        - Also needs some cleanup to make sure it doesn't stay there.
        - Also does upload_latest_model truly take the latest?
        - What if latest locally and in the cloud are outdated? It needs to go in the right folder.
        """
        # TODO: implement as soon as platform is out of beta mode
        pytest.skip("Test not yet implemented")

        # Make dummy production data
        issue_dir, kwargs = make_production_data(
            self.sync_dir, team="Demo", machine="Charger 75kW", service="Diagnostics"
        )
        # Upload
        sync = PlatformSynchronizer()
        sync.upload_latest_model(issue_dir, **kwargs)


class TestAmploPlatformAPI:
    """
    Tests for the class `AmploPlatformAPI`.
    """

    @classmethod
    def setup_class(cls):
        cls.api = AmploPlatformAPI.from_os_env()

    def test_request(self):
        # A valid request should return a Response object
        out = self.api.request("get", "models", params={"team": "Demo"})
        assert isinstance(out, requests.Response)
        # An invalid request should raise an error
        with pytest.raises(requests.HTTPError):
            self.api.request("get", "does_not_exist")

    def test_list_models(self):
        params = {"team": "Demo"}
        expected_out = self.api.request("get", "models", params=params).json()
        actual_out = self.api.list_models(**params)
        assert expected_out == actual_out

    def test_list_trainings(self):
        params = {"team": "Demo", "service": "Diagnostics", "category": "Diagnostics"}
        expected_out = self.api.request("get", "trainings", params=params).json()
        actual_out = self.api.list_trainings(**params)
        assert expected_out == actual_out

    def test_upload_trainings(self):
        pytest.skip("Test not yet implemented. We need a dummy training_id for that.")


def test_format_version():
    """
    Test the function `_format_version`.
    """

    # Case 1: integer
    assert _format_version(1) == "v1"

    # Case 2: string
    assert _format_version("1") == "v1"
    assert _format_version("v1") == "v1"

    # Case 3: bad string format
    with pytest.raises(ValueError):
        _format_version("abc")


def test_upload_model():
    pytest.skip("Test not yet implemented. We need a dummy training_id for that.")

    # api = AmploPlatformAPI.from_os_env()
    upload_model(**{"this": "line", "will": "fail"})



#  Copyright (c) 2022 by Amplo.

import warnings

import pandas as pd
from sklearn.feature_selection import r_regression  # pearson coefficient
from sklearn.preprocessing import LabelEncoder

from amplo.utils.util import clean_feature_name

__all__ = [
    "influx_query_to_df",
    "check_dataframe_quality",
    "check_pearson_correlation",
]


def influx_query_to_df(result):
    df = []
    for table in result:
        parsed_records = []
        for record in table.records:
            parsed_records.append((record.get_time(), record.get_value()))
        df.append(pd.DataFrame(parsed_records, columns=["ts", record.get_field()]))
    return pd.concat(df).set_index("ts").groupby(level=0).sum()


def check_dataframe_quality(data: pd.DataFrame) -> bool:
    if data.isna().any().any():
        warnings.warn("Data contains NaN.")
    elif data.isnull().any().any():
        warnings.warn("Data contains null.")
    elif (data.dtypes == object).any().any():
        warnings.warn("Data contains dtype 'object', which is ambiguous.")
    elif (data.dtypes == str).any().any():
        warnings.warn("Data contains dtype 'str', which is ambiguous.")
    elif data.max().max() > 1e38 or data.min().min() < -1e38:
        warnings.warn("Data contains values larger than float32 (1e38).")
    else:
        return True
    return False


def check_pearson_correlation(features: pd.DataFrame, labels: pd.Series) -> bool:
    if labels.dtype == "object":
        labels = LabelEncoder().fit_transform(labels)
    pearson_corr = r_regression(features, labels)
    if abs(pearson_corr).mean() > 0.5:
        return False
    else:
        return True
